# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 28.7 `interp` Chapter.

``interp`` - Interpolation

This module is concerned with the interpolation of a function of one or
more variables.
When provided with the value of the function (and possibly one or more
of its lowest-order derivatives) at each of a number of values of the
variable(s), the NAG Library functions provide either an interpolating
function or an interpolated value.
For some of the interpolating functions, there are supporting NAG
Library functions to evaluate, differentiate or integrate them.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01intro.html
"""

# NAG Copyright 2017-2022.

from __future__ import absolute_import

import collections as _collections

from math import sqrt

from ..base import interp as _b_interp
from ..base import machine as _b_machine
from ..base import utils as _b_utils
from ..library import _utils

Dim1AitkenReturnData = _collections.namedtuple(
    'Dim1AitkenReturnData',
    ['a', 'b', 'c'],
)

def dim1_aitken(a, b, x):
    u"""
    Interpolated values, Aitken's technique, unequally spaced data, one
    variable.

    ``dim1_aitken`` interpolates a function of one variable at a given
    point x from a table of function values y_i evaluated at equidistant
    or non-equidistant points x_i, for i = 1,2,...,n+1, using Aitken's
    technique of successive linear interpolations.

    For full information please refer to the NAG Library document for
    e01aa

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01aaf.html

    Parameters
    ----------
    a : float, array-like, shape (n+1)
        `a`[i-1] must contain the x-component of the ith data point,
        x_i, for i = 1,2,...,n+1.

    b : float, array-like, shape (n+1)
        `b`[i-1] must contain the y-component (function value) of the
        ith data point, y_i, for i = 1,2,...,n+1.

    x : float
        The point x at which the interpolation is required. Note that
        `x` may lie outside the interval defined by the minimum and
        maximum values in `a`, in which case an extrapolated value will
        be computed; extrapolated results should be treated with
        considerable caution since there is no information on the
        behaviour of the function outside the defined interval.

    Returns
    -------
    a : float, ndarray, shape (n+1)
        `a`[i-1] contains the value x_i-x, for i = 1,2,...,n+1.

    b : float, ndarray, shape (n+1)
        The contents of `b` are unspecified.

    c : float, ndarray, shape (n*(n+1)/2)
            `c`[0],...,`c`[n-1] contain the first set of linear
            interpolations,

            `c`[n],...,`c`[2*n-2] contain the second set of linear
            interpolations,

            `c`[2n-1],...,`c`[3*n-4] contain the third set of linear
            interpolations,

            ...

            `c`[n*(n+1)/2-1] contains the interpolated function value at
            the point x.

    Raises
    ------
    NagValueError
        (`errno` 6)
            On entry, error in parameter n.

            Constraint: n > 0.

    Notes
    -----
    ``dim1_aitken`` interpolates a function of one variable at a given
    point x from a table of values x_i and y_i, for i = 1,2,...,n+1
    using Aitken's method (see Fröberg (1970)).
    The intermediate values of linear interpolations are stored to
    enable an estimate of the accuracy of the results to be made.

    References
    ----------
    Fröberg, C E, 1970, `Introduction to Numerical Analysis`,
    Addison--Wesley
    """

    fun_name = 'naginterfaces.library.interp.dim1_aitken'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py*(_n_py+1)//2,),
    )
    _b_interp.dim1_aitken(
        _a,
        _b,
        _c,
        _n,
        x,
    )
    _a_py = _a
    _b_py = _b
    _c_py = _c
    _return_data = Dim1AitkenReturnData(
        _a_py,
        _b_py,
        _c_py,
    )
    return _return_data

Dim1EverettReturnData = _collections.namedtuple(
    'Dim1EverettReturnData',
    ['a', 'g'],
)

def dim1_everett(p, a):
    u"""
    Interpolated values, Everett's formula, equally spaced data, one
    variable.

    ``dim1_everett`` interpolates a function of one variable at a given
    point x from a table of function values evaluated at equidistant
    points, using Everett's formula.

    For full information please refer to the NAG Library document for
    e01ab

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01abf.html

    Parameters
    ----------
    p : float
        The point p at which the interpolated function value is
        required, i.e., p = (x-x_0)/h with -1.0 < p < 1.0.

    a : float, array-like, shape (2*n)
        `a`[i-1] must be set to the function value y_{i-n}, for i =
        1,2,...,2n.

    Returns
    -------
    a : float, ndarray, shape (2*n)
        The contents of `a` are unspecified.

    g : float, ndarray, shape (2*n+1)
        The array contains

        [table omitted]

        The interpolated function value y_p is stored in `g`[2n].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` <= 1.0.

        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` >= -1.0.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n > 0.

    Notes
    -----
    ``dim1_everett`` interpolates a function of one variable at a given
    point

        x = x_0+ph,

    where -1 <= p <= 1 and h is the interval of differencing, from a
    table of values x_m = x_0+mh and y_m where m =
    -(n-1),-(n-2),...,-1,0,1,...,n.
    The formula used is that of Fröberg (1970), neglecting the remainder
    term:

        y_p = sum_r = 0^n-1((1-p+r)/(2r+1))delta^(2r)y_0+sum_r =
        0^n-1((p+r)/(2r+1))delta^(2r)y_1.

    The values of delta^(2r)y_0 and delta^(2r)y_1 are stored on exit
    from the function in addition to the interpolated function value
    y_p.

    References
    ----------
    Fröberg, C E, 1970, `Introduction to Numerical Analysis`,
    Addison--Wesley
    """

    fun_name = 'naginterfaces.library.interp.dim1_everett'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]//2, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _g = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2*_n_py+1,),
    )
    _b_interp.dim1_everett(
        _n,
        p,
        _a,
        _g,
    )
    _a_py = _a
    _g_py = _g
    _return_data = Dim1EverettReturnData(
        _a_py,
        _g_py,
    )
    return _return_data

Dim1ChebReturnData = _collections.namedtuple(
    'Dim1ChebReturnData',
    ['a', 'wrk', 'iwrk'],
)

def dim1_cheb(xmin, xmax, x, y, ip, lwrk, liwrk, itmin=0, itmax=0):
    u"""
    Interpolating functions, polynomial interpolant, data may include
    derivative values, one variable.

    ``dim1_cheb`` constructs the Chebyshev series representation of a
    polynomial interpolant to a set of data which may contain derivative
    values.

    For full information please refer to the NAG Library document for
    e01ae

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01aef.html

    Parameters
    ----------
    xmin : float
        The lower and upper end points, respectively, of the interval
        [x_min,x_max]. If they are not determined by your problem, it is
        recommended that they be set respectively to the smallest and
        largest values among the x_i.

    xmax : float
        The lower and upper end points, respectively, of the interval
        [x_min,x_max]. If they are not determined by your problem, it is
        recommended that they be set respectively to the smallest and
        largest values among the x_i.

    x : float, array-like, shape (m)
        The value of x_i, for i = 1,2,...,m. The `x`[i-1] need not be
        ordered.

    y : float, array-like, shape (n)
        The given values of the dependent variable, and derivatives, as
        follows:

        The first p_1+1 elements contain y_1,y_1^(1),...,y_1^(p_1) in
        that order.

        The next p_2+1 elements contain y_2,y_2^(1),...,y_2^(p_2) in
        that order.

         ...

        The last p_m+1 elements contain y_m,y_m^(1),...,y_m^(p_m) in
        that order.

    ip : int, array-like, shape (m)
        p_i, the order of the highest-order derivative whose value is
        given at x_i, for i = 1,2,...,m. If the value of y only is given
        for some x_i then the corresponding value of `ip`[i-1] must be
        zero.

    lwrk : int
        The dimension of the array `wrk`.

    liwrk : int
        The dimension of the array `iwrk`.

    itmin : int, optional
        Respectively the minimum and maximum number of iterations to be
        performed by the function (for full details see [Further
        Comments]). Setting `itmin` and/or `itmax` negative or zero
        invokes default value(s) of 2 and/or 10, respectively.

        The default values will be satisfactory for most problems, but
        occasionally significant improvement will result from using
        higher values.

    itmax : int, optional
        Respectively the minimum and maximum number of iterations to be
        performed by the function (for full details see [Further
        Comments]). Setting `itmin` and/or `itmax` negative or zero
        invokes default value(s) of 2 and/or 10, respectively.

        The default values will be satisfactory for most problems, but
        occasionally significant improvement will result from using
        higher values.

    Returns
    -------
    a : float, ndarray, shape (n)
        `a`[i] contains the coefficient a_i in the Chebyshev series
        representation of q(x), for i = 0,...,n-1.

    wrk : float, ndarray, shape (`lwrk`)
        Used as workspace, but see also [Further Comments].

    iwrk : int, ndarray, shape (`liwrk`)
        Used as workspace, but see also [Further Comments].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `liwrk` is too small. `liwrk` = *<value>*. Minimum
            possible dimension: *<value>*.

        (`errno` 1)
            On entry, `lwrk` is too small. `lwrk` = *<value>*. Minimum
            possible dimension: *<value>*.

        (`errno` 1)
            On entry, n = *<value>* and m+`ip`[0]+`ip`[1]+...+`ip`[m-1]
            = *<value>*.

            Constraint: n = m+`ip`[0]+`ip`[1]+...+`ip`[m-1].

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 2)
            On entry, I = *<value>* and `ip`[I-1] = *<value>*.

            Constraint: `ip`[I-1] >= 0.

        (`errno` 3)
            On entry, I = *<value>*, J = *<value>* and `x`[I-1] =
            *<value>*.

            Constraint: `x`[I-1] != `x`[J-1].

        (`errno` 3)
            On entry, I = *<value>*, `x`[I-1] = *<value>*, `xmin` =
            *<value>* and `xmax` = *<value>*.

            Constraint: `xmin` <= `x`[I-1] <= `xmax`.

        (`errno` 3)
            On entry, `xmin` = *<value>* and `xmax` = *<value>*.

            Constraint: `xmin` < `xmax`.

        (`errno` 5)
            The computation has been terminated because the iterative
            process appears to be diverging.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            Not all the performance indices are less than eight times
            the machine precision, although `itmax` iterations have been
            performed. A more accurate solution may possibly be obtained
            by increasing `itmax` and recalling the function.

    Notes
    -----
    Let m distinct values x_i of an independent variable x be given,
    with x_min <= x_i <= x_max, for i = 1,2,...,m.
    For each value x_i, suppose that the value y_i of the dependent
    variable y together with the first p_i derivatives of y with respect
    to x are given.
    Each p_i must, therefore, be a non-negative integer, with the total
    number of interpolating conditions, n, equal to m+sum_i = 1^mp_i.

    ``dim1_cheb`` calculates the unique polynomial q(x) of degree n-1
    (or less) which is such that q^(k)(x_i) = y_i^(k), for k =
    0,1,...,p_i, for i = 1,2,...,m.
    Here q^(0)(x_i) means q(x_i).
    This polynomial is represented in Chebyshev series form in the
    normalized variable x_, as follows:

        q(x) = 1/2a_0 T_0(x_)+a_1 T_1(x_)+...+a_{n-1}T_{n-1}(x_),

    where

        x_= (2x-x_min-x_max)/(x_max-x_min)

    so that -1 <= x_<= 1 for x in the interval x_min to x_max, and where
    T_i(x_) is the Chebyshev polynomial of the first kind of degree i
    with argument x_.

    (The polynomial interpolant can subsequently be evaluated for any
    value of x in the given range by using ``fit.dim1_cheb_eval2``.
    Chebyshev series representations of the derivative(s) and
    integral(s) of q(x) may be obtained by (repeated) use of
    ``fit.dim1_cheb_deriv`` and ``fit.dim1_cheb_integ``.)

    The method used consists first of constructing a divided-difference
    table from the normalized x_ values and the given values of y and
    its derivatives with respect to x_.
    The Newton form of q(x) is then obtained from this table, as
    described in Huddleston (1974) and Krogh (1970), with the
    modification described in [Further Comments].
    The Newton form of the polynomial is then converted to Chebyshev
    series form as described in [Conversion to Chebyshev Form].

    Since the errors incurred by these stages can be considerable, a
    form of iterative refinement is used to improve the solution.
    This refinement is particularly useful when derivatives of rather
    high order are given in the data.
    In reasonable examples, the refinement will usually terminate with a
    certain accuracy criterion satisfied by the polynomial (see
    [Accuracy]).
    In more difficult examples, the criterion may not be satisfied and
    refinement will continue until the maximum number of iterations (as
    specified by the input argument `itmax`) is reached.

    In extreme examples, the iterative process may diverge (even though
    the accuracy criterion is satisfied): if a certain divergence
    criterion is satisfied, the process terminates at once.
    In all cases the function returns the 'best' polynomial achieved
    before termination.
    For the definition of 'best' and details of iterative refinement and
    termination criteria, see [Further Comments].

    References
    ----------
    Huddleston, R E, 1974, `CDC 6600 routines for the interpolation of
    data and of data with derivatives`, SLL-74-0214, Sandia Laboratories
    (Reprint)

    Krogh, F T, 1970, `Efficient algorithms for polynomial interpolation
    and numerical differentiation`, Math. Comput. (24), 185--190
    """

    fun_name = 'naginterfaces.library.interp.dim1_cheb'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ip = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ip'},
        ip,
        copy=False,
    )
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _lwrk = _b_utils._EngineIntScalarType(
        lwrk, {'fun_name': fun_name, 'entity': 'lwrk'},
    ).to_ctypes()
    _lwrk_py = _b_utils._EngineIntScalarType.to_py(_lwrk)
    _wrk = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_lwrk_py,),
    )
    _liwrk = _b_utils._EngineIntScalarType(
        liwrk, {'fun_name': fun_name, 'entity': 'liwrk'},
    ).to_ctypes()
    _liwrk_py = _b_utils._EngineIntScalarType.to_py(_liwrk)
    _iwrk = _b_utils._EngineIntArrayType.empty_ndarray(
        (_liwrk_py,),
    )
    try:
        _b_interp.dim1_cheb(
            _m,
            xmin,
            xmax,
            _x,
            _y,
            _ip,
            _n,
            itmin,
            itmax,
            _a,
            _wrk,
            _lwrk,
            _iwrk,
            _liwrk,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _wrk_py = _wrk
    _iwrk_py = _iwrk
    _return_data = Dim1ChebReturnData(
        _a_py,
        _wrk_py,
        _iwrk_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

Dim1SplineReturnData = _collections.namedtuple(
    'Dim1SplineReturnData',
    ['lamda', 'c'],
)

def dim1_spline(x, y):
    u"""
    Interpolating functions, cubic spline interpolant, one variable.

    ``dim1_spline`` determines a cubic spline interpolant to a given set
    of data.

    For full information please refer to the NAG Library document for
    e01ba

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01baf.html

    Parameters
    ----------
    x : float, array-like, shape (m)
        `x`[i-1] must be set to x_i, the ith data value of the
        independent variable x, for i = 1,2,...,m.

    y : float, array-like, shape (m)
        `y`[i-1] must be set to y_i, the ith data value of the dependent
        variable y, for i = 1,2,...,m.

    Returns
    -------
    lamda : float, ndarray, shape (m+4)
        The value of lambda_i, the ith knot, for i = 1,2,...,m+4.

    c : float, ndarray, shape (m)
        The coefficient c_i of the B-spline N_i(x), for i = 1,2,...,m.
        The remaining elements of the array are not used.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, lck = *<value>* and m = *<value>*.

            Constraint: lck >= m+4.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 4.

        (`errno` 2)
            On entry, I = *<value>*, `x`[I-1] = *<value>* and `x`[I-2] =
            *<value>*.

            Constraint: `x`[I-1] > `x`[I-2].

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim1_spline`` determines a cubic spline s(x), defined in the range
    x_1 <= x <= x_m, which interpolates (passes exactly through) the set
    of data points (x_i,y_i), for i = 1,2,...,m, where m >= 4 and x_1 <
    x_2 < ... < x_m.
    Unlike some other spline interpolation algorithms, derivative end
    conditions are not imposed.
    The spline interpolant chosen has m-4 interior knots
    lambda_5,lambda_6,...,lambda_m, which are set to the values of
    x_3,x_4,...,x_{m-2} respectively.
    This spline is represented in its B-spline form (see Cox (1975)):

        s(x) = sum_i = 1^mc_i N_i(x),

    where N_i(x) denotes the normalized B-spline of degree 3, defined
    upon the knots lambda_i,lambda_{i+1},...,lambda_{i+4}, and c_i
    denotes its coefficient, whose value is to be determined by the
    function.

    The use of B-splines requires eight additional knots lambda_1,
    lambda_2, lambda_3, lambda_4, lambda_{m+1}, lambda_{m+2},
    lambda_{m+3} and lambda_{m+4} to be specified; ``dim1_spline`` sets
    the first four of these to x_1 and the last four to x_m.

    The algorithm for determining the coefficients is as described in
    Cox (1975) except that QR factorization is used instead of LU
    decomposition.
    The implementation of the algorithm involves setting up appropriate
    information for the related function ``fit.dim1_spline_knots``
    followed by a call of that function. (See ``fit.dim1_spline_knots``
    for further details.)

    Values of the spline interpolant, or of its derivatives or definite
    integral, can subsequently be computed as detailed in [Further
    Comments].

    References
    ----------
    Cox, M G, 1975, `An algorithm for spline interpolation`, J. Inst.
    Math. Appl. (15), 95--108

    Cox, M G, 1977, `A survey of numerical methods for data and function
    approximation`, The State of the Art in Numerical Analysis, (ed D A
    H Jacobs), 627--668, Academic Press
    """

    fun_name = 'naginterfaces.library.interp.dim1_spline'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lck = _b_utils._EngineIntScalarType(
        _m_py+4, {'fun_name': fun_name, 'entity': '_lck'},
    ).to_ctypes()
    _lamda = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py+4,),
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py+4,),
    )
    _b_interp.dim1_spline(
        _m,
        _x,
        _y,
        _lamda,
        _c,
        _lck,
    )
    _lamda_py = _lamda
    _c_py = _c[:_m_py]
    _return_data = Dim1SplineReturnData(
        _lamda_py,
        _c_py,
    )
    return _return_data

def dim1_monotonic(x, f):
    u"""
    Interpolating functions, monotonicity-preserving, piecewise cubic
    Hermite, one variable.

    ``dim1_monotonic`` computes a monotonicity-preserving piecewise
    cubic Hermite interpolant to a set of data points.

    For full information please refer to the NAG Library document for
    e01be

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01bef.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        `x`[r-1] must be set to x_r, the rth value of the independent
        variable (abscissa), for r = 1,2,...,n.

    f : float, array-like, shape (n)
        `f`[r-1] must be set to f_r, the rth value of the dependent
        variable (ordinate), for r = 1,2,...,n.

    Returns
    -------
    d : float, ndarray, shape (n)
        Estimates of derivatives at the data points. `d`[r-1] contains
        the derivative at `x`[r-1].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 2)
            On entry, r = *<value>*, `x`[r-2] = *<value>* and `x`[r-1] =
            *<value>*.

            Constraint: `x`[r-2] < `x`[r-1] for all r.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim1_monotonic`` estimates first derivatives at the set of data
    points (x_r,f_r), for r = 1,2,...,n, which determine a piecewise
    cubic Hermite interpolant to the data, that preserves monotonicity
    over ranges where the data points are monotonic.
    If the data points are only piecewise monotonic, the interpolant
    will have an extremum at each point where monotonicity switches
    direction.
    The estimates of the derivatives are computed by a formula due to
    Brodlie, which is described in Fritsch and Butland (1984), with
    suitable changes at the boundary points.

    The function is derived from function PCHIM in Fritsch (1982).

    Values of the computed interpolant, and of its first derivative and
    definite integral, can subsequently be computed by calling
    ``dim1_monotonic_eval``, ``dim1_monotonic_deriv`` and
    ``dim1_monotonic_intg``, as described in [Further Comments].

    References
    ----------
    Fritsch, F N, 1982, `PCHIP final specifications`, Report UCID-30194,
    Lawrence Livermore National Laboratory

    Fritsch, F N and Butland, J, 1984, `A method for constructing local
    monotone piecewise cubic interpolants`, SIAM J. Sci. Statist.
    Comput. (5), 300--304
    """

    fun_name = 'naginterfaces.library.interp.dim1_monotonic'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _b_interp.dim1_monotonic(
        _n,
        _x,
        _f,
        _d,
    )
    _d_py = _d
    return _d_py

Dim1MonotonicEvalReturnData = _collections.namedtuple(
    'Dim1MonotonicEvalReturnData',
    ['pf'],
)

def dim1_monotonic_eval(x, f, d, px):
    u"""
    Interpolated values, interpolant computed by ``dim1_monotonic``,
    function only, one variable.

    ``dim1_monotonic_eval`` evaluates a piecewise cubic Hermite
    interpolant at a set of points.

    For full information please refer to the NAG Library document for
    e01bf

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01bff.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        n, `x`, `f` and `d` must be unchanged from the previous call of
        ``dim1_monotonic``.

    f : float, array-like, shape (n)
        n, `x`, `f` and `d` must be unchanged from the previous call of
        ``dim1_monotonic``.

    d : float, array-like, shape (n)
        n, `x`, `f` and `d` must be unchanged from the previous call of
        ``dim1_monotonic``.

    px : float, array-like, shape (m)
        The m values of x at which the interpolant is to be evaluated.

    Returns
    -------
    pf : float, ndarray, shape (m)
        `pf`[i-1] contains the value of the interpolant evaluated at the
        point `px`[i-1], for i = 1,2,...,m.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 2)
            On entry, r = *<value>*, `x`[r-2] = *<value>* and `x`[r-1] =
            *<value>*.

            Constraint: `x`[r-2] < `x`[r-1] for all r.

        (`errno` 3)
            On entry, m = *<value>*.

            Constraint: m >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            Warning -- some points in array `px` lie outside the range
            `x`[0]...`x`[n-1]. Values at these points are unreliable
            because computed by extrapolation.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim1_monotonic_eval`` evaluates a piecewise cubic Hermite
    interpolant, as computed by ``dim1_monotonic``, at the points
    `px`[i-1], for i = 1,2,...,m.
    If any point lies outside the interval from `x`[0] to `x`[n-1], a
    value is extrapolated from the nearest extreme cubic, and a warning
    is returned.

    The function is derived from function PCHFE in Fritsch (1982).

    References
    ----------
    Fritsch, F N, 1982, `PCHIP final specifications`, Report UCID-30194,
    Lawrence Livermore National Laboratory
    """

    fun_name = 'naginterfaces.library.interp.dim1_monotonic_eval'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _px = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'px'},
        px,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_px, {'fun_name': fun_name, 'entity': 'px'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _pf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    try:
        _b_interp.dim1_monotonic_eval(
            _n,
            _x,
            _f,
            _d,
            _m,
            _px,
            _pf,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _pf_py = _pf
    _return_data = Dim1MonotonicEvalReturnData(
        _pf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _pf_py

Dim1MonotonicDerivReturnData = _collections.namedtuple(
    'Dim1MonotonicDerivReturnData',
    ['pf', 'pd'],
)

def dim1_monotonic_deriv(x, f, d, px):
    u"""
    Interpolated values, interpolant computed by ``dim1_monotonic``,
    function and first derivative, one variable.

    ``dim1_monotonic_deriv`` evaluates a piecewise cubic Hermite
    interpolant and its first derivative at a set of points.

    For full information please refer to the NAG Library document for
    e01bg

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01bgf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        n, `x`, `f` and `d` must be unchanged from the previous call of
        ``dim1_monotonic``.

    f : float, array-like, shape (n)
        n, `x`, `f` and `d` must be unchanged from the previous call of
        ``dim1_monotonic``.

    d : float, array-like, shape (n)
        n, `x`, `f` and `d` must be unchanged from the previous call of
        ``dim1_monotonic``.

    px : float, array-like, shape (m)
        The m values of x at which the interpolant is to be evaluated.

    Returns
    -------
    pf : float, ndarray, shape (m)
        `pf`[i-1] contains the value of the interpolant evaluated at the
        point `px`[i-1], for i = 1,2,...,m.

    pd : float, ndarray, shape (m)
        `pd`[i-1] contains the first derivative of the interpolant
        evaluated at the point `px`[i-1], for i = 1,2,...,m.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 2)
            On entry, r = *<value>*, `x`[r-2] = *<value>* and `x`[r-1] =
            *<value>*.

            Constraint: `x`[r-2] < `x`[r-1] for all r.

        (`errno` 3)
            On entry, m = *<value>*.

            Constraint: m >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            Warning -- some points in array `px` lie outside the range
            `x`[0]...`x`[n-1]. Values at these points are unreliable
            because computed by extrapolation.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim1_monotonic_deriv`` evaluates a piecewise cubic Hermite
    interpolant, as computed by ``dim1_monotonic``, at the points
    `px`[i-1], for i = 1,2,...,m.
    The first derivatives at the points are also computed.
    If any point lies outside the interval from `x`[0] to `x`[n-1],
    values of the interpolant and its derivative are extrapolated from
    the nearest extreme cubic, and a warning is returned.

    If values of the interpolant only, and not of its derivative, are
    required, ``dim1_monotonic_eval`` should be used.

    The function is derived from function PCHFD in Fritsch (1982).

    References
    ----------
    Fritsch, F N, 1982, `PCHIP final specifications`, Report UCID-30194,
    Lawrence Livermore National Laboratory
    """

    fun_name = 'naginterfaces.library.interp.dim1_monotonic_deriv'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _px = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'px'},
        px,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_px, {'fun_name': fun_name, 'entity': 'px'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _pf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _pd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    try:
        _b_interp.dim1_monotonic_deriv(
            _n,
            _x,
            _f,
            _d,
            _m,
            _px,
            _pf,
            _pd,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _pf_py = _pf
    _pd_py = _pd
    _return_data = Dim1MonotonicDerivReturnData(
        _pf_py,
        _pd_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

Dim1MonotonicIntgReturnData = _collections.namedtuple(
    'Dim1MonotonicIntgReturnData',
    ['pint'],
)

def dim1_monotonic_intg(x, f, d, a, b):
    u"""
    Interpolated values, interpolant computed by ``dim1_monotonic``,
    definite integral, one variable.

    ``dim1_monotonic_intg`` evaluates the definite integral of a
    piecewise cubic Hermite interpolant over the interval [a,b].

    For full information please refer to the NAG Library document for
    e01bh

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01bhf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        n, `x`, `f` and `d` must be unchanged from the previous call of
        ``dim1_monotonic``.

    f : float, array-like, shape (n)
        n, `x`, `f` and `d` must be unchanged from the previous call of
        ``dim1_monotonic``.

    d : float, array-like, shape (n)
        n, `x`, `f` and `d` must be unchanged from the previous call of
        ``dim1_monotonic``.

    a : float
        The interval [a,b] over which integration is to be performed.

    b : float
        The interval [a,b] over which integration is to be performed.

    Returns
    -------
    pint : float
        The value of the definite integral of the interpolant over the
        interval [a,b].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 2)
            On entry, r = *<value>*, `x`[r-2] = *<value>* and `x`[r-1] =
            *<value>*.

            Constraint: `x`[r-2] < `x`[r-1] for all r.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Warning -- either `a` or `b` is outside the range
            `x`[0]...`x`[n-1]. The result has been computed by
            extrapolation and is unreliable. `a` = *<value>* `b` =
            *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim1_monotonic_intg`` evaluates the definite integral of a
    piecewise cubic Hermite interpolant, as computed by
    ``dim1_monotonic``, over the interval [a,b].

    If either a or b lies outside the interval from `x`[0] to `x`[n-1]
    computation of the integral involves extrapolation and a warning is
    returned.

    The function is derived from function PCHIA in Fritsch (1982).

    References
    ----------
    Fritsch, F N, 1982, `PCHIP final specifications`, Report UCID-30194,
    Lawrence Livermore National Laboratory
    """

    fun_name = 'naginterfaces.library.interp.dim1_monotonic_intg'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalar = _b_interp.dim1_monotonic_intg(
            _n,
            _x,
            _f,
            _d,
            a,
            b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _pint_py = _return_scalar
    _return_data = Dim1MonotonicIntgReturnData(
        _pint_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _pint_py

def dim1_monconv_disc(negfor, yfor, x, y, lam=0.2):
    u"""
    Interpolating variables, monotone convex Hagan--West procedure, one
    variable.

    ``dim1_monconv_disc`` computes, for a given set of data points, the
    forward values and other values required for monotone convex
    interpolation as defined in Hagan and West (2008).
    This form of interpolation is particularly suited to the
    construction of yield curves in Financial Mathematics but can be
    applied to any data where it is desirable to preserve both
    monotonicity and convexity.

    For full information please refer to the NAG Library document for
    e01ce

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01cef.html

    Parameters
    ----------
    negfor : bool
        Determines whether or not to allow negative forward rates.

        `negfor` = True

            Negative forward rates are permitted.

        `negfor` = False

            Forward rates calculated must be non-negative.

    yfor : bool
        Determines whether the array `y` contains values, y, or discrete
        forward rates f^d.

        `yfor` = True

            `y` contains the discrete forward rates f_i^d, for i =
            1,2,...,n.

        `yfor` = False

            `y` contains the values y_i, for i = 1,2,...,n.

    x : float, array-like, shape (n)
        x, the (possibly unordered) set of data points.

    y : float, array-like, shape (n)
        If `yfor` = True, the discrete forward rates f_i^d corresponding
        to the data points x_i, for i = 1,2,...,n.

        If `yfor` = False, the data values y_i corresponding to the data
        points x_i, for i = 1,2,...,n.

    lam : float, optional
        lambda, the amelioration (smoothing) parameter. Forward rates
        are first computed using (2) and then, if lambda > 0, a limiting
        filter is applied which depends on neighbouring discrete forward
        values. This filter has a smoothing effect on the curve that
        increases with lambda.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 2)
            On entry, `lam` = *<value>*.

            Constraint: 0.0 <= `lam` <= 1.0.

        (`errno` 3)
            On entry, `x` contains duplicate data points.

    Notes
    -----
    ``dim1_monconv_disc`` computes, for a set of data points, (x_i,y_i),
    for i = 1,2,...,n, the discrete forward rates, f_i^d, and the
    instantaneous forward rates, f_i, which are used in a monotone
    convex interpolation method that attempts to preserve both the
    monotonicity and the convexity of the original data.
    The monotone convex interpolation method is due to Hagan and West
    and is described in Hagan and West (2006), Hagan and West (2008) and
    West (2011).

    The discrete forward rates are defined simply, for ordered data, by

        f_1^d = y_1; ; f_i^d = (x_i y_i-x_{i-1}y_{i-1})/(x_i-x_{i-1}),
        for i = 2,3,...,n.

    The discrete forward rates, if pre-computed, may be supplied instead
    of y, in which case the original values y are computed using the
    inverse of (1).

    The data points x_i need not be ordered on input (though y_i must
    correspond to x_i); a set of ordered and scaled values xi_i are
    calculated from x_i and stored.

    In its simplest form, the instantaneous forward rates, f_i, at the
    data points are computed as linear interpolations of the f_i^d:

        f_i = (x_i-x_{i-1})/(x_{i+1}-x_{i-1}) f_i+1^d +
        (x_{i+1}-x_i)/(x_{i+1}-x_{i-1}) f_i^d, for i = 2,3,...,n-1 ;f_1
        = f_2^d - 1/2 (f_2-f_2^d); f_n = f_n^d - 1/2 (f_{n-1}-f_n^d).

    If it is required, as a constraint, that these values should never
    be negative then a limiting filter is applied to f as described in
    Section 3.6 of West (2011).

    An ameliorated (smoothed) form of this linear interpolation for the
    forward rates is implemented using the amelioration (smoothing)
    parameter lambda.
    For lambda = 0, equation (2) is used (with possible post-process
    filtering); for 0 < lambda <= 1, the ameliorated method described
    fully in West (2011) is used.

    The values computed by ``dim1_monconv_disc`` are used by
    ``dim1_monconv_eval`` to compute, for a given value x^, the monotone
    convex interpolated (or extrapolated) value y^(x^) and the
    corresponding instantaneous forward rate f; the curve gradient at x^
    can be derived as y' = (f-y^)/x^ for x^!= 0.

    References
    ----------
    Hagan, P S and West, G, 2006, `Interpolation methods for curve
    construction`, Applied Mathematical Finance (13(2)), 89--129

    Hagan, P S and West, G, 2008, `Methods for constructing a yield
    curve`, WILLMOTT Magazine (May), 70--81

    West, G, 2011, `The monotone convex method of interpolation`,
    Financial Modelling Agency
    """

    fun_name = 'naginterfaces.library.interp.dim1_monconv_disc'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_interp.dim1_monconv_disc(
        _n,
        lam,
        negfor,
        yfor,
        _x,
        _y,
    )
    _comm_py = _return_scalar
    return _comm_py

Dim1MonconvEvalReturnData = _collections.namedtuple(
    'Dim1MonconvEvalReturnData',
    ['val', 'fwd'],
)

def dim1_monconv_eval(x, comm):
    u"""
    Interpolated values, variables computed by ``dim1_monconv_disc``,
    monotone convex Hagan--West procedure, one variable.

    ``dim1_monconv_eval`` evaluates a monotonic convex interpolant at a
    set of points.

    For full information please refer to the NAG Library document for
    e01cf

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01cff.html

    Parameters
    ----------
    x : float, array-like, shape (m)
        x, the points at which the interpolant is to be evaluated.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``dim1_monconv_disc``.

    Returns
    -------
    val : float, ndarray, shape (m)
        The values of the interpolant at x.

    fwd : float, ndarray, shape (m)
        The values of the forward rates at x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Either ``dim1_monconv_disc`` was not called first or the
            communication array has become corrupted.

    Notes
    -----
    ``dim1_monconv_eval`` evaluates a monotonic convex interpolant, as
    setup by ``dim1_monconv_disc``, at the points x.
    The function is derived from the work of Hagan and West and is
    described in Hagan and West (2006), Hagan and West (2008) and West
    (2011).

    References
    ----------
    Hagan, P S and West, G, 2006, `Interpolation methods for curve
    construction`, Applied Mathematical Finance (13(2)), 89--129

    Hagan, P S and West, G, 2008, `Methods for constructing a yield
    curve`, WILLMOTT Magazine (May), 70--81

    West, G, 2011, `The monotone convex method of interpolation`,
    Financial Modelling Agency
    """

    fun_name = 'naginterfaces.library.interp.dim1_monconv_eval'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _val = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _fwd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _b_interp.dim1_monconv_eval(
        _m,
        _x,
        _val,
        _fwd,
        comm,
    )
    _val_py = _val
    _fwd_py = _fwd
    _return_data = Dim1MonconvEvalReturnData(
        _val_py,
        _fwd_py,
    )
    return _return_data

Dim2SplineGridReturnData = _collections.namedtuple(
    'Dim2SplineGridReturnData',
    ['px', 'py', 'lamda', 'mu', 'c'],
)

def dim2_spline_grid(x, y, f):
    u"""
    Interpolating functions, fitting bicubic spline, data on rectangular
    grid.

    ``dim2_spline_grid`` computes a bicubic spline interpolating surface
    through a set of data values, given on a rectangular grid in the x-y
    plane.

    For full information please refer to the NAG Library document for
    e01da

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01daf.html

    Parameters
    ----------
    x : float, array-like, shape (mx)
        `x`[q-1] and `y`[r-1] must contain x_q, for q = 1,2,...,m_x, and
        y_r, for r = 1,2,...,m_y, respectively.

    y : float, array-like, shape (my)
        `x`[q-1] and `y`[r-1] must contain x_q, for q = 1,2,...,m_x, and
        y_r, for r = 1,2,...,m_y, respectively.

    f : float, array-like, shape (mx*my)
        `f`[m_y*(q-1)+r-1] must contain f_{q,r}, for r = 1,2,...,m_y,
        for q = 1,2,...,m_x.

    Returns
    -------
    px : int
        `px` and `py` contain m_x+4 and m_y+4, the total number of knots
        of the computed spline with respect to the x and y variables,
        respectively.

    py : int
        `px` and `py` contain m_x+4 and m_y+4, the total number of knots
        of the computed spline with respect to the x and y variables,
        respectively.

    lamda : float, ndarray, shape (mx+4)
        `lamda` contains the complete set of knots lambda_i associated
        with the x variable

    mu : float, ndarray, shape (my+4)
        `mu` contains the complete set of knots mu_i associated with the
        y variable

    c : float, ndarray, shape (mx*my)
        The coefficients of the spline interpolant. `c`[m_y*(i-1)+j-1]
        contains the coefficient c_{ij} described in [Notes].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, my = *<value>*.

            Constraint: my >= 4.

        (`errno` 1)
            On entry, mx = *<value>*.

            Constraint: mx >= 4.

        (`errno` 2)
            On entry, the `x` or the `y` mesh points are not in strictly
            ascending order.

        (`errno` 3)
            An intermediate set of linear equations is singular -- the
            data is too ill-conditioned to compute B-spline
            coefficients.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim2_spline_grid`` determines a bicubic spline interpolant to the
    set of data points (x_q,y_r,f_{q,r}), for r = 1,2,...,m_y, for q =
    1,2,...,m_x.
    The spline is given in the B-spline representation

        s(x,y) = sum_i = 1^m_xsum_j = 1^m_yc_{ij}M_i(x)N_j(y),

    such that

        s(x_q,y_r) = f_{q,r},

    where M_i(x) and N_j(y) denote normalized cubic B-splines, the
    former defined on the knots lambda_i to lambda_{i+4} and the latter
    on the knots mu_j to mu_{j+4}, and the c_{ij} are the spline
    coefficients.
    These knots, as well as the coefficients, are determined by the
    function, which is derived from the function B2IRE in Anthony `et
    al.` (1982).
    The method used is described in [Further Comments].

    For further information on splines, see Hayes and Halliday (1974)
    for bicubic splines and de Boor (1972) for normalized B-splines.

    Values and derivatives of the computed spline can subsequently be
    computed by calling ``fit.dim2_spline_evalv``,
    ``fit.dim2_spline_evalm`` or ``fit.dim2_spline_derivm`` as described
    in [Evaluation of Computed Spline].

    References
    ----------
    Anthony, G T, Cox, M G and Hayes, J G, 1982, `DASL -- Data
    Approximation Subroutine Library`, National Physical Laboratory

    Cox, M G, 1975, `An algorithm for spline interpolation`, J. Inst.
    Math. Appl. (15), 95--108

    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62

    Hayes, J G and Halliday, J, 1974, `The least squares fitting of
    cubic spline surfaces to general data sets`, J. Inst. Math. Appl.
    (14), 89--103
    """

    fun_name = 'naginterfaces.library.interp.dim2_spline_grid'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _mx = _b_utils._EngineIntScalarType(
        _mx_py, {'fun_name': fun_name, 'entity': '_mx'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _my_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _my = _b_utils._EngineIntScalarType(
        _my_py, {'fun_name': fun_name, 'entity': '_my'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lamda = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_mx_py+4,),
    )
    _mu = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_my_py+4,),
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_mx_py*_my_py,),
    )
    _return_scalars = _b_interp.dim2_spline_grid(
        _mx,
        _my,
        _x,
        _y,
        _f,
        _lamda,
        _mu,
        _c,
    )
    _px_py = _return_scalars.px
    _py_py = _return_scalars.py
    _lamda_py = _lamda
    _mu_py = _mu
    _c_py = _c
    _return_data = Dim2SplineGridReturnData(
        _px_py,
        _py_py,
        _lamda_py,
        _mu_py,
        _c_py,
    )
    return _return_data

def dim2_triangulate(x, y):
    u"""
    Triangulation of two-dimensional scattered grid, method of Renka and
    Cline.

    ``dim2_triangulate`` generates a triangulation for a given set of
    two-dimensional points using the method of Renka and Cline.

    For full information please refer to the NAG Library document for
    e01ea

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01eaf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The x coordinates of the n data points.

    y : float, array-like, shape (n)
        The y coordinates of the n data points.

    Returns
    -------
    triang : int, ndarray, shape (7*n)
        A data structure defining the computed triangulation, in a form
        suitable for passing to ``dim2_triang_bary_eval``. Details of
        how the triangulation is encoded in `triang` are given in
        [Further Comments]. These details are most likely to be of use
        when plotting the computed triangulation.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 3.

        (`errno` 2)
            On entry, all the (x,y) pairs are collinear.

    Notes
    -----
    ``dim2_triangulate`` creates a Thiessen triangulation with a given
    set of two-dimensional data points as nodes.
    This triangulation will be as equiangular as possible (Cline and
    Renka (1984)).
    See Renka and Cline (1984) for more detailed information on the
    algorithm, a development of that by Lawson (1977).
    The code is derived from Renka (1984).

    The computed triangulation is returned in a form suitable for
    passing to ``dim2_triang_bary_eval`` which, for a set of nodal
    function values, computes interpolated values at a set of points.

    References
    ----------
    Cline, A K and Renka, R L, 1984, `A storage-efficient method for
    construction of a Thiessen triangulation`, Rocky Mountain J. Math.
    (14), 119--139

    Lawson, C L, 1977, `Software for C^1 surface interpolation`,
    Mathematical Software III, (ed J R Rice), 161--194, Academic Press

    Renka, R L, 1984, `Algorithm 624: triangulation and interpolation of
    arbitrarily distributed points in the plane`, ACM Trans. Math.
    Software (10), 440--442

    Renka, R L and Cline, A K, 1984, `A triangle-based C^1 interpolation
    method`, Rocky Mountain J. Math. (14), 223--237
    """

    fun_name = 'naginterfaces.library.interp.dim2_triangulate'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _triang = _b_utils._EngineIntArrayType.empty_ndarray(
        (7*_n_py,),
    )
    _b_interp.dim2_triangulate(
        _n,
        _x,
        _y,
        _triang,
    )
    _triang_py = _triang
    return _triang_py

Dim2TriangBaryEvalReturnData = _collections.namedtuple(
    'Dim2TriangBaryEvalReturnData',
    ['pf'],
)

def dim2_triang_bary_eval(x, y, f, triang, px, py):
    u"""
    Barycentric interpolation on function values provided on a
    two-dimensional scattered grid.

    ``dim2_triang_bary_eval`` performs barycentric interpolation, at a
    given set of points, using a set of function values on a scattered
    grid and a triangulation of that grid computed by
    ``dim2_triangulate``.

    For full information please refer to the NAG Library document for
    e01eb

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01ebf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The coordinates of the rth data point, (x_r,y_r), for r =
        1,2,...,n. `x` and `y` must be unchanged from the previous call
        of ``dim2_triangulate``.

    y : float, array-like, shape (n)
        The coordinates of the rth data point, (x_r,y_r), for r =
        1,2,...,n. `y` and `y` must be unchanged from the previous call
        of ``dim2_triangulate``.

    f : float, array-like, shape (n)
        The function values f_r at (x_r,y_r), for r = 1,2,...,n.

    triang : int, array-like, shape (7*n)
        The triangulation computed by the previous call of
        ``dim2_triangulate``. See [Further Comments] for details of how
        the triangulation used is encoded in `triang`.

    px : float, array-like, shape (m)
        The coordinates (px_i,py_i), for i = 1,2,...,m, at which
        interpolated function values are sought.

    py : float, array-like, shape (m)
        The coordinates (px_i,py_i), for i = 1,2,...,m, at which
        interpolated function values are sought.

    Returns
    -------
    pf : float, ndarray, shape (m)
        The interpolated values F(px_i,py_i), for i = 1,2,...,m.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 3.

        (`errno` 2)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 3)
            On entry, the triangulation information held in the array
            `triang` does not specify a valid triangulation of the data
            points. `triang` has been corrupted since the call to
            ``dim2_triangulate``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            At least one evaluation point lies outside the nodal
            triangulation. For each such point the value returned in
            `pf` is that corresponding to a node on the closest boundary
            line segment.

    Notes
    -----
    ``dim2_triang_bary_eval`` takes as input a set of scattered data
    points (x_r,y_r,f_r), for r = 1,2,...,n, and a Thiessen
    triangulation of the (x_r,y_r) computed by ``dim2_triangulate``, and
    interpolates at a set of points (px_i,py_i), for i = 1,2,...,m.

    If the ith interpolation point (px_i,py_i) is equal to (x_r,y_r) for
    some value of r, the returned value will be equal to f_r; otherwise
    a barycentric transformation will be used to calculate the
    interpolant.

    For each point (px_i,py_i), a triangle is sought which contains the
    point; the vertices of the triangle and f_r values at the vertices
    are then used to compute the value F(px_i,py_i).

    If any interpolation point lies outside the triangulation defined by
    the input arguments, the returned value is the value provided, f_s,
    at the closest node (x_s,y_s).

    ``dim2_triang_bary_eval`` must only be called after a call to
    ``dim2_triangulate``.

    References
    ----------
    Cline, A K and Renka, R L, 1984, `A storage-efficient method for
    construction of a Thiessen triangulation`, Rocky Mountain J. Math.
    (14), 119--139

    Lawson, C L, 1977, `Software for C^1 surface interpolation`,
    Mathematical Software III, (ed J R Rice), 161--194, Academic Press

    Renka, R L, 1984, `Algorithm 624: triangulation and interpolation of
    arbitrarily distributed points in the plane`, ACM Trans. Math.
    Software (10), 440--442

    Renka, R L and Cline, A K, 1984, `A triangle-based C^1 interpolation
    method`, Rocky Mountain J. Math. (14), 223--237
    """

    fun_name = 'naginterfaces.library.interp.dim2_triang_bary_eval'

    _px = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'px'},
        px,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_px, {'fun_name': fun_name, 'entity': 'px'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _triang = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'triang'},
        triang,
        copy=False,
    )
    _py = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'py'},
        py,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _pf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    try:
        _b_interp.dim2_triang_bary_eval(
            _m,
            _n,
            _x,
            _y,
            _f,
            _triang,
            _px,
            _py,
            _pf,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _pf_py = _pf
    _return_data = Dim2TriangBaryEvalReturnData(
        _pf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _pf_py

Dim1RatnlReturnData = _collections.namedtuple(
    'Dim1RatnlReturnData',
    ['m', 'a', 'u'],
)

def dim1_ratnl(x, f):
    u"""
    Interpolating functions, rational interpolant, one variable.

    ``dim1_ratnl`` produces, from a set of function values and
    corresponding abscissae, the coefficients of an interpolating
    rational function expressed in continued fraction form.

    For full information please refer to the NAG Library document for
    e01ra

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01raf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        `x`[i-1] must be set to the value of the ith data abscissa, x_i,
        for i = 1,2,...,n.

    f : float, array-like, shape (n)
        `f`[i-1] must be set to the value of the data ordinate, f_i,
        corresponding to x_i, for i = 1,2,...,n.

    Returns
    -------
    m : int
        m, the number of terms in the continued fraction representation
        of R(x).

    a : float, ndarray, shape (n)
        `a`[j-1] contains the value of the parameter a_j in R(x), for j
        = 1,2,...,m. The remaining elements of `a`, if any, are set to
        zero.

    u : float, ndarray, shape (n)
        `u`[j-1] contains the value of the parameter u_j in R(x), for j
        = 1,2,...,m-1. The u_j are a permuted subset of the elements of
        `x`. The remaining n-m+1 locations contain a permutation of the
        remaining x_i, which can be ignored.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 2)
            On entry, `x`[I-1] is very close to `x`[J-1]: I = *<value>*,
            `x`[I-1] = *<value>*, J = *<value>* and `x`[J-1] =
            *<value>*.

        (`errno` 3)
            A continued fraction of the required form does not exist.

    Notes
    -----
    ``dim1_ratnl`` produces the parameters of a rational function R(x)
    which assumes prescribed values f_i at prescribed values x_i of the
    independent variable x, for i = 1,2,...,n.
    More specifically, ``dim1_ratnl`` determines the parameters a_j, for
    j = 1,2,...,m and u_j, for j = 1,2,...,m-1, in the continued
    fraction

        R(x) = a_1+R_m(x)

    where

        R_i(x) = a_{m-i+2}(x-u_{m-i+1})/(1+R_{i-1}(x)), for i =
        m,m-1,...,2,

    and

        R_1(x) = 0,

    such that R(x_i) = f_i, for i = 1,2,...,n.
    The value of m in (1) is determined by the function; normally m = n.
    The values of u_j form a reordered subset of the values of x_i and
    their ordering is designed to ensure that a representation of the
    form (1) is determined whenever one exists.

    The subsequent evaluation of (1) for given values of x can be
    carried out using ``dim1_ratnl_eval``.

    The computational method employed in ``dim1_ratnl`` is the
    modification of the Thacher--Tukey algorithm described in
    Graves--Morris and Hopkins (1981).

    References
    ----------
    Graves--Morris, P R and Hopkins, T R, 1981, `Reliable rational
    interpolation`, Numer. Math. (36), 111--128
    """

    fun_name = 'naginterfaces.library.interp.dim1_ratnl'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _u = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalar = _b_interp.dim1_ratnl(
        _n,
        _x,
        _f,
        _a,
        _u,
    )
    _m_py = _return_scalar
    _a_py = _a
    _u_py = _u
    _return_data = Dim1RatnlReturnData(
        _m_py,
        _a_py,
        _u_py,
    )
    return _return_data

def dim1_ratnl_eval(a, u, x):
    u"""
    Interpolated values, evaluate rational interpolant computed by
    ``dim1_ratnl``, one variable.

    ``dim1_ratnl_eval`` evaluates continued fractions of the form
    produced by ``dim1_ratnl``.

    For full information please refer to the NAG Library document for
    e01rb

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01rbf.html

    Parameters
    ----------
    a : float, array-like, shape (m)
        `a`[j-1] must be set to the value of the parameter a_j in the
        continued fraction, for j = 1,2,...,m.

    u : float, array-like, shape (m)
        `u`[j-1] must be set to the value of the parameter u_j in the
        continued fraction, for j = 1,2,...,m-1. (The element `u`[m-1]
        is not used).

    x : float
        The value of x at which the continued fraction is to be
        evaluated.

    Returns
    -------
    f : float
        The value of the continued fraction corresponding to the value
        of x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `x` corresponds to a pole of R(x), or is very close. `x` =
            *<value>*.

    Notes
    -----
    ``dim1_ratnl_eval`` evaluates the continued fraction

        R(x) = a_1+R_m(x)

    where

        R_i(x) = a_{m-i+2}(x-u_{m-i+1})/(1+R_{i-1}(x)), for i =
        m,m-1,...,2.

    and

        R_1(x) = 0

    for a prescribed value of x. ``dim1_ratnl_eval`` is intended to be
    used to evaluate the continued fraction representation (of an
    interpolatory rational function) produced by ``dim1_ratnl``.

    References
    ----------
    Graves--Morris, P R and Hopkins, T R, 1981, `Reliable rational
    interpolation`, Numer. Math. (36), 111--128
    """

    fun_name = 'naginterfaces.library.interp.dim1_ratnl_eval'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_interp.dim1_ratnl_eval(
        _m,
        _a,
        _u,
        x,
    )
    _f_py = _return_scalar
    return _f_py

Dim2ScatReturnData = _collections.namedtuple(
    'Dim2ScatReturnData',
    ['triang', 'grads'],
)

def dim2_scat(x, y, f):
    u"""
    Interpolating functions, method of Renka and Cline, two variables.

    ``dim2_scat`` generates a two-dimensional surface interpolating a
    set of scattered data points, using the method of Renka and Cline.

    For full information please refer to the NAG Library document for
    e01sa

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01saf.html

    Parameters
    ----------
    x : float, array-like, shape (m)
        The coordinates of the rth data point, for r = 1,2,...,m. The
        data points are accepted in any order, but see [Further
        Comments].

    y : float, array-like, shape (m)
        The coordinates of the rth data point, for r = 1,2,...,m. The
        data points are accepted in any order, but see [Further
        Comments].

    f : float, array-like, shape (m)
        The coordinates of the rth data point, for r = 1,2,...,m. The
        data points are accepted in any order, but see [Further
        Comments].

    Returns
    -------
    triang : int, ndarray, shape (7*m)
        A data structure defining the computed triangulation, in a form
        suitable for passing to ``dim2_scat_eval``.

    grads : float, ndarray, shape (2, m)
        The estimated partial derivatives at the nodes, in a form
        suitable for passing to ``dim2_scat_eval``. The derivatives at
        node r with respect to x and y are contained in `grads`[0,r-1]
        and `grads`[1,r-1] respectively, for r = 1,2,...,m.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 3.

        (`errno` 2)
            All nodes are collinear. There is no unique solution.

        (`errno` 3)
            On entry, (`x`[I-1],`y`[I-1]) = (`x`[J-1],`y`[J-1]), for I,J
            = *<value>**<value>*, `x`[I-1], `y`[I-1] =
            *<value>**<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim2_scat`` constructs an interpolating surface F(x, y) through a
    set of m scattered data points (x_r,y_r,f_r), for r = 1,2,...,m,
    using a method due to Renka and Cline.
    In the (x,y) plane, the data points must be distinct.
    The constructed surface is continuous and has continuous first
    derivatives.

    The method involves firstly creating a triangulation with all the
    (x,y) data points as nodes, the triangulation being as nearly
    equiangular as possible (see Cline and Renka (1984)).
    Then gradients in the x- and y-directions are estimated at node r,
    for r = 1,2,...,m, as the partial derivatives of a quadratic
    function of x and y which interpolates the data value f_r, and which
    fits the data values at nearby nodes (those within a certain
    distance chosen by the algorithm) in a weighted least squares sense.
    The weights are chosen such that closer nodes have more influence
    than more distant nodes on derivative estimates at node r.
    The computed partial derivatives, with the f_r values, at the three
    nodes of each triangle define a piecewise polynomial surface of a
    certain form which is the interpolant on that triangle.
    See Renka and Cline (1984) for more detailed information on the
    algorithm, a development of that by Lawson (1977).
    The code is derived from Renka (1984).

    The interpolant F(x, y) can subsequently be evaluated at any point
    (x,y) inside or outside the domain of the data by a call to
    ``dim2_scat_eval``. Points outside the domain are evaluated by
    extrapolation.

    References
    ----------
    Cline, A K and Renka, R L, 1984, `A storage-efficient method for
    construction of a Thiessen triangulation`, Rocky Mountain J. Math.
    (14), 119--139

    Lawson, C L, 1977, `Software for C^1 surface interpolation`,
    Mathematical Software III, (ed J R Rice), 161--194, Academic Press

    Renka, R L, 1984, `Algorithm 624: triangulation and interpolation of
    arbitrarily distributed points in the plane`, ACM Trans. Math.
    Software (10), 440--442

    Renka, R L and Cline, A K, 1984, `A triangle-based C^1 interpolation
    method`, Rocky Mountain J. Math. (14), 223--237
    """

    fun_name = 'naginterfaces.library.interp.dim2_scat'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _triang = _b_utils._EngineIntArrayType.empty_ndarray(
        (7*_m_py,),
    )
    _grads = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2, _m_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _b_interp.dim2_scat(
        _m,
        _x,
        _y,
        _f,
        _triang,
        _grads,
    )
    _triang_py = _triang
    _grads_py = _grads
    _return_data = Dim2ScatReturnData(
        _triang_py,
        _grads_py,
    )
    return _return_data

Dim2ScatEvalReturnData = _collections.namedtuple(
    'Dim2ScatEvalReturnData',
    ['ist', 'pf'],
)

def dim2_scat_eval(x, y, f, triang, grads, px, py, ist=1):
    u"""
    Interpolated values, evaluate interpolant computed by ``dim2_scat``,
    two variables.

    ``dim2_scat_eval`` evaluates at a given point the two-dimensional
    interpolant function computed by ``dim2_scat``.

    For full information please refer to the NAG Library document for
    e01sb

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01sbf.html

    Parameters
    ----------
    x : float, array-like, shape (m)
        `x` must be unchanged from the previous call of ``dim2_scat``

    y : float, array-like, shape (m)
        `y` must be unchanged from the previous call of ``dim2_scat``

    f : float, array-like, shape (m)
        `f` must be unchanged from the previous call of ``dim2_scat``

    triang : int, array-like, shape (7*m)
        `triang` must be unchanged from the previous call of
        ``dim2_scat``

    grads : float, array-like, shape (2, m)
        `grads` must be unchanged from the previous call of
        ``dim2_scat``

    px : float
        The point (px,py) at which the interpolant is to be evaluated.

    py : float
        The point (px,py) at which the interpolant is to be evaluated.

    ist : int, optional
        The index of the starting node in the search for a triangle
        containing the point (px,py). On the first call to
        ``dim2_scat_eval``, `ist` must be set to 1. For efficiency on
        subsequent calls to ``dim2_scat_eval`` an updated value of `ist`
        as returned by ``dim2_scat_eval`` may be supplied instead. An
        input value outside the range 1 <= `ist` <= m will be treated as
        1.

    Returns
    -------
    ist : int
        The index of one of the vertices of the triangle containing the
        point (px,py).

    pf : float
        The value of the interpolant evaluated at the point (px,py).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 3.

        (`errno` 2)
            On entry, `triang` does not contain a valid data point
            triangulation; `triang` may have been corrupted since the
            call to ``dim2_scat``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Warning -- the evaluation point (*<value>*,*<value>*) lies
            outside the triangulation boundary. The returned value was
            computed by extrapolation.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``dim2_scat_eval`` takes as input the arguments defining the
    interpolant F(x, y) of a set of scattered data points (x_r,y_r,f_r),
    for r = 1,2,...,m, as computed by ``dim2_scat``, and evaluates the
    interpolant at the point (px,py).

    If (px,py) is equal to (x_r,y_r) for some value of r, the returned
    value will be equal to f_r.

    If (px,py) is not equal to (x_r,y_r) for any r, the derivatives in
    `grads` will be used to compute the interpolant.
    A triangle is sought which contains the point (px,py), and the
    vertices of the triangle along with the partial derivatives and f_r
    values at the vertices are used to compute the value F(px, py).
    If the point (px,py) lies outside the triangulation defined by the
    input arguments, the returned value is obtained by extrapolation.
    In this case, the interpolating function `f` is extended linearly
    beyond the triangulation boundary.
    The method is described in more detail in Renka and Cline (1984) and
    the code is derived from Renka (1984).

    ``dim2_scat_eval`` must only be called after a call to
    ``dim2_scat``.

    References
    ----------
    Renka, R L, 1984, `Algorithm 624: triangulation and interpolation of
    arbitrarily distributed points in the plane`, ACM Trans. Math.
    Software (10), 440--442

    Renka, R L and Cline, A K, 1984, `A triangle-based C^1 interpolation
    method`, Rocky Mountain J. Math. (14), 223--237
    """

    fun_name = 'naginterfaces.library.interp.dim2_scat_eval'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _triang = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'triang'},
        triang,
        copy=False,
    )
    _grads = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'grads'},
        grads,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalars = _b_interp.dim2_scat_eval(
            _m,
            _x,
            _y,
            _f,
            _triang,
            _grads,
            px,
            py,
            ist,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ist_py = _return_scalars.ist
    _pf_py = _return_scalars.pf
    _return_data = Dim2ScatEvalReturnData(
        _ist_py,
        _pf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

Dim2ScatShepReturnData = _collections.namedtuple(
    'Dim2ScatShepReturnData',
    ['iq', 'rq'],
)

def dim2_scat_shep(x, y, f, nw, nq):
    u"""
    Interpolating functions, modified Shepard's method, two variables.

    ``dim2_scat_shep`` generates a two-dimensional interpolant to a set
    of scattered data points, using a modified Shepard method.

    For full information please refer to the NAG Library document for
    e01sg

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01sgf.html

    Parameters
    ----------
    x : float, array-like, shape (m)
        The Cartesian coordinates of the data points (x_r,y_r), for r =
        1,2,...,m.

    y : float, array-like, shape (m)
        The Cartesian coordinates of the data points (x_r,y_r), for r =
        1,2,...,m.

    f : float, array-like, shape (m)
        `f`[r-1] must be set to the data value f_r, for r = 1,2,...,m.

    nw : int
        The number N_w of data points that determines each radius of
        influence R_w, appearing in the definition of each of the
        weights w_r, for r = 1,2,...,m (see [Notes]). Note that R_w is
        different for each weight. If `nw` <= 0 the default value `nw` =
        min(19, m-1) is used instead.

    nq : int
        The number N_q of data points to be used in the least squares
        fit for coefficients defining the nodal functions q_r(x,y) (see
        [Notes]). If `nq` <= 0 the default value `nq` = min(13, m-1) is
        used instead.

    Returns
    -------
    iq : int, ndarray, shape (2*m+1)
        Integer data defining the interpolant Q(x,y).

    rq : float, ndarray, shape (6*m+5)
        Real data defining the interpolant Q(x,y).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, lrq is too small: lrq = *<value>*.

        (`errno` 1)
            On entry, liq is too small: liq = *<value>*.

        (`errno` 1)
            On entry, `nw` = *<value>* and m = *<value>*.

            Constraint: `nw` <= min(40, m-1).

        (`errno` 1)
            On entry, `nq` = *<value>* and m = *<value>*.

            Constraint: `nq` <= min(40, m-1).

        (`errno` 1)
            On entry, `nq` = *<value>*.

            Constraint: `nq` <= 0 or `nq` >= 5.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 6.

        (`errno` 2)
            There are duplicate nodes in the dataset.
            (`x`[I-1],`y`[I-1]) = (`x`[J-1],`y`[J-1]), for I = *<value>*
            and J = *<value>*. The interpolant cannot be derived.

        (`errno` 3)
            All nodes are collinear. There is no unique solution.

    Notes
    -----
    ``dim2_scat_shep`` constructs a smooth function Q(x,y) which
    interpolates a set of m scattered data points (x_r,y_r,f_r), for r =
    1,2,...,m, using a modification of Shepard's method.
    The surface is continuous and has continuous first partial
    derivatives.

    The basic Shepard (1968) method interpolates the input data with the
    weighted mean

        Q(x,y) = sum_r = 1^mw_r(x,y)q_r/(sum_r = 1^mw_r(x,y)),

    where q_r = f_r, w_r(x,y) = 1/d_r^2 and d_r^2 = (x-x_r)^2+(y-y_r)^2.

    The basic method is global in that the interpolated value at any
    point depends on all the data, but this function uses a modification
    (see Franke and Nielson (1980) and Renka (1988a)), whereby the
    method becomes local by adjusting each w_r(x,y) to be zero outside a
    circle with centre (x_r,y_r) and some radius R_w.
    Also, to improve the performance of the basic method, each q_r above
    is replaced by a function q_r(x,y), which is a quadratic fitted by
    weighted least squares to data local to (x_r,y_r) and forced to
    interpolate (x_r,y_r,f_r).
    In this context, a point (x,y) is defined to be local to another
    point if it lies within some distance R_q of it.
    Computation of these quadratics constitutes the main work done by
    this function.

    The efficiency of the function is further enhanced by using a cell
    method for nearest neighbour searching due to Bentley and Friedman
    (1979).

    The radii R_w and R_q are chosen to be just large enough to include
    N_w and N_q data points, respectively, for user-supplied constants
    N_w and N_q.
    Default values of these arguments are provided by the function, and
    advice on alternatives is given in [Further Comments].

    This function is derived from the function QSHEP2 described by Renka
    (1988b).

    Values of the interpolant Q(x,y) generated by this function, and its
    first partial derivatives, can subsequently be evaluated for points
    in the domain of the data by a call to ``dim2_scat_shep_eval``.

    References
    ----------
    Bentley, J L and Friedman, J H, 1979, `Data structures for range
    searching`, ACM Comput. Surv. (11), 397--409

    Franke, R and Nielson, G, 1980, `Smooth interpolation of large sets
    of scattered data`, Internat. J. Num. Methods Engrg. (15),
    1691--1704

    Renka, R J, 1988, `Algorithm 660: QSHEP2D: Quadratic Shepard method
    for bivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 149--150

    Renka, R J, 1988, `Multivariate interpolation of large sets of
    scattered data`, ACM Trans. Math. Software (14), 139--148

    Shepard, D, 1968, `A two-dimensional interpolation function for
    irregularly spaced data`, Proc. 23rd Nat. Conf. ACM, 517--523,
    Brandon/Systems Press Inc., Princeton
    """

    fun_name = 'naginterfaces.library.interp.dim2_scat_shep'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _liq = _b_utils._EngineIntScalarType(
        2*_m_py+1, {'fun_name': fun_name, 'entity': '_liq'},
    ).to_ctypes()
    _iq = _b_utils._EngineIntArrayType.empty_ndarray(
        (2*_m_py+1,),
    )
    _lrq = _b_utils._EngineIntScalarType(
        6*_m_py+5, {'fun_name': fun_name, 'entity': '_lrq'},
    ).to_ctypes()
    _rq = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (6*_m_py+5,),
    )
    _b_interp.dim2_scat_shep(
        _m,
        _x,
        _y,
        _f,
        nw,
        nq,
        _iq,
        _liq,
        _rq,
        _lrq,
    )
    _iq_py = _iq
    _rq_py = _rq
    _return_data = Dim2ScatShepReturnData(
        _iq_py,
        _rq_py,
    )
    return _return_data

Dim2ScatShepEvalReturnData = _collections.namedtuple(
    'Dim2ScatShepEvalReturnData',
    ['q', 'qx', 'qy'],
)

def dim2_scat_shep_eval(x, y, f, iq, rq, u, v):
    u"""
    Interpolated values, evaluate interpolant computed by
    ``dim2_scat_shep``, function and first derivatives, two variables.

    ``dim2_scat_shep_eval`` evaluates the two-dimensional interpolating
    function generated by ``dim2_scat_shep`` and its first partial
    derivatives.

    For full information please refer to the NAG Library document for
    e01sh

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01shf.html

    Parameters
    ----------
    x : float, array-like, shape (m)
        m, `x`, `y` and `f` must be the same values as were supplied in
        the preceding call to ``dim2_scat_shep``.

    y : float, array-like, shape (m)
        m, `x`, `y` and `f` must be the same values as were supplied in
        the preceding call to ``dim2_scat_shep``.

    f : float, array-like, shape (m)
        m, `x`, `y` and `f` must be the same values as were supplied in
        the preceding call to ``dim2_scat_shep``.

    iq : int, array-like, shape (2*m+1)
        Must be unchanged from the value returned from a previous call
        to ``dim2_scat_shep``.

    rq : float, array-like, shape (6*m+5)
        Must be unchanged from the value returned from a previous call
        to ``dim2_scat_shep``.

    u : float, array-like, shape (n)
        The evaluation points (u_i,v_i), for i = 1,2,...,n.

    v : float, array-like, shape (n)
        The evaluation points (u_i,v_i), for i = 1,2,...,n.

    Returns
    -------
    q : float, ndarray, shape (n)
        The values of the interpolant at (u_i,v_i), for i = 1,2,...,n.
        If any of these evaluation points lie outside the region of
        definition of the interpolant the corresponding entries in `q`
        are set to an extrapolated approximation, and
        ``dim2_scat_shep_eval`` returns with `errno` = 3.

    qx : float, ndarray, shape (n)
        The values of the partial derivatives of the interpolant Q(x,y)
        at (u_i,v_i), for i = 1,2,...,n. If any of these evaluation
        points lie outside the region of definition of the interpolant,
        the corresponding entries in `qx` and `qy` are set to
        extrapolated approximations to the partial derivatives, and
        ``dim2_scat_shep_eval`` returns with `errno` = 3.

    qy : float, ndarray, shape (n)
        The values of the partial derivatives of the interpolant Q(x,y)
        at (u_i,v_i), for i = 1,2,...,n. If any of these evaluation
        points lie outside the region of definition of the interpolant,
        the corresponding entries in `qx` and `qy` are set to
        extrapolated approximations to the partial derivatives, and
        ``dim2_scat_shep_eval`` returns with `errno` = 3.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, lrq is too small: lrq = *<value>*.

        (`errno` 1)
            On entry, liq is too small: liq = *<value>*.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 6.

        (`errno` 2)
            On entry, values in `rq` appear to be invalid. Check that
            `rq` has not been corrupted between calls to
            ``dim2_scat_shep`` and ``dim2_scat_shep_eval``.

        (`errno` 2)
            On entry, values in `iq` appear to be invalid. Check that
            `iq` has not been corrupted between calls to
            ``dim2_scat_shep`` and ``dim2_scat_shep_eval``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, at least one evaluation point lies outside the
            region of definition of the interpolant. At such points the
            corresponding values in `q` and `qx` contain extrapolated
            approximations. Points should be evaluated one by one to
            identify extrapolated values.

    Notes
    -----
    ``dim2_scat_shep_eval`` takes as input the interpolant Q(x,y) of a
    set of scattered data points (x_r,y_r,f_r), for r = 1,2,...,m, as
    computed by ``dim2_scat_shep``, and evaluates the interpolant and
    its first partial derivatives at the set of points (u_i,v_i), for i
    = 1,2,...,n.

    ``dim2_scat_shep_eval`` must only be called after a call to
    ``dim2_scat_shep``.

    This function is derived from the function QS2GRD described by Renka
    (1988).

    References
    ----------
    Renka, R J, 1988, `Algorithm 660: QSHEP2D: Quadratic Shepard method
    for bivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 149--150
    """

    fun_name = 'naginterfaces.library.interp.dim2_scat_shep_eval'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _iq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iq'},
        iq,
        copy=False,
    )
    _liq_py = _b_utils._EngineIntArrayType.get_shape(_iq, {'fun_name': fun_name, 'entity': 'iq'}, exp_rank=1)[0]
    _liq = _b_utils._EngineIntScalarType(
        _liq_py, {'fun_name': fun_name, 'entity': '_liq'},
    ).to_ctypes()
    _rq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rq'},
        rq,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lrq_py = _b_utils._EngineFloat64ArrayType.get_shape(_rq, {'fun_name': fun_name, 'entity': 'rq'}, exp_rank=1)[0]
    _lrq = _b_utils._EngineIntScalarType(
        _lrq_py, {'fun_name': fun_name, 'entity': '_lrq'},
    ).to_ctypes()
    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _v = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'v'},
        v,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _q = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _qx = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _qy = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_interp.dim2_scat_shep_eval(
            _m,
            _x,
            _y,
            _f,
            _iq,
            _liq,
            _rq,
            _lrq,
            _n,
            _u,
            _v,
            _q,
            _qx,
            _qy,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _q_py = _q
    _qx_py = _qx
    _qy_py = _qy
    _return_data = Dim2ScatShepEvalReturnData(
        _q_py,
        _qx_py,
        _qy_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

Dim3ScatShepReturnData = _collections.namedtuple(
    'Dim3ScatShepReturnData',
    ['iq', 'rq'],
)

def dim3_scat_shep(x, y, z, f, nw, nq):
    u"""
    Interpolating functions, modified Shepard's method, three variables.

    ``dim3_scat_shep`` generates a three-dimensional interpolant to a
    set of scattered data points, using a modified Shepard method.

    For full information please refer to the NAG Library document for
    e01tg

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01tgf.html

    Parameters
    ----------
    x : float, array-like, shape (m)
        `x`[r-1], `y`[r-1], `z`[r-1] must be set to the Cartesian
        coordinates of the data point (x_r,y_r,z_r), for r = 1,2,...,m.

    y : float, array-like, shape (m)
        `x`[r-1], `y`[r-1], `z`[r-1] must be set to the Cartesian
        coordinates of the data point (x_r,y_r,z_r), for r = 1,2,...,m.

    z : float, array-like, shape (m)
        `x`[r-1], `y`[r-1], `z`[r-1] must be set to the Cartesian
        coordinates of the data point (x_r,y_r,z_r), for r = 1,2,...,m.

    f : float, array-like, shape (m)
        `f`[r-1] must be set to the data value f_r, for r = 1,2,...,m.

    nw : int
        The number N_w of data points that determines each radius of
        influence R_w, appearing in the definition of each of the
        weights w_r, for r = 1,2,...,m (see [Notes]). Note that R_w is
        different for each weight. If `nw` <= 0 the default value `nw` =
        min(32, m-1) is used instead.

    nq : int
        The number N_q of data points to be used in the least squares
        fit for coefficients defining the nodal functions q_r(x,y,z)
        (see [Notes]). If `nq` <= 0 the default value `nq` = min(17,
        m-1) is used instead.

    Returns
    -------
    iq : int, ndarray, shape (2*m+1)
        Integer data defining the interpolant Q(x,y,z).

    rq : float, ndarray, shape (10*m+7)
        Real data defining the interpolant Q(x,y,z).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, lrq is too small: lrq = *<value>*.

        (`errno` 1)
            On entry, liq is too small: liq = *<value>*.

        (`errno` 1)
            On entry, `nw` = *<value>* and m = *<value>*.

            Constraint: `nw` <= min(40, m-1).

        (`errno` 1)
            On entry, `nq` = *<value>* and m = *<value>*.

            Constraint: `nq` <= min(40, m-1).

        (`errno` 1)
            On entry, `nq` = *<value>*.

            Constraint: `nq` <= 0 or `nq` >= 9.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 10.

        (`errno` 2)
            There are duplicate nodes in the dataset.
            (`x`[I-1],`y`[I-1],`z`[I-1]) = (`x`[J-1],`y`[J-1],`z`[J-1])
            for: I = *<value>* and J = *<value>*. The interpolant cannot
            be derived.

        (`errno` 3)
            All nodes are coplanar. There is no unique solution.

    Notes
    -----
    ``dim3_scat_shep`` constructs a smooth function Q(x,y,z) which
    interpolates a set of m scattered data points (x_r,y_r,z_r,f_r), for
    r = 1,2,...,m, using a modification of Shepard's method.
    The surface is continuous and has continuous first partial
    derivatives.

    The basic Shepard method, which is a generalization of the
    two-dimensional method described in Shepard (1968), interpolates the
    input data with the weighted mean

        Q(x,y,z) = sum_r = 1^mw_r(x,y,z)q_r/(sum_r = 1^mw_r(x,y,z)),

    where

        q_r = f_r and w_r(x,y,z) = 1/d_r^2 and d_r^2 =
        (x-x_r)^2+(y-y_r)^2+(z-z_r)^2.

    The basic method is global in that the interpolated value at any
    point depends on all the data, but this function uses a modification
    (see Franke and Nielson (1980) and Renka (1988a)), whereby the
    method becomes local by adjusting each w_r(x,y,z) to be zero outside
    a sphere with centre (x_r,y_r,z_r) and some radius R_w.
    Also, to improve the performance of the basic method, each q_r above
    is replaced by a function q_r(x,y,z), which is a quadratic fitted by
    weighted least squares to data local to (x_r,y_r,z_r) and forced to
    interpolate (x_r,y_r,z_r,f_r).
    In this context, a point (x,y,z) is defined to be local to another
    point if it lies within some distance R_q of it.
    Computation of these quadratics constitutes the main work done by
    this function.

    The efficiency of the function is further enhanced by using a cell
    method for nearest neighbour searching due to Bentley and Friedman
    (1979).

    The radii R_w and R_q are chosen to be just large enough to include
    N_w and N_q data points, respectively, for user-supplied constants
    N_w and N_q.
    Default values of these arguments are provided by the function, and
    advice on alternatives is given in [Further Comments].

    This function is derived from the function QSHEP3 described by Renka
    (1988b).

    Values of the interpolant Q(x,y,z) generated by this function, and
    its first partial derivatives, can subsequently be evaluated for
    points in the domain of the data by a call to
    ``dim3_scat_shep_eval``.

    References
    ----------
    Bentley, J L and Friedman, J H, 1979, `Data structures for range
    searching`, ACM Comput. Surv. (11), 397--409

    Franke, R and Nielson, G, 1980, `Smooth interpolation of large sets
    of scattered data`, Internat. J. Num. Methods Engrg. (15),
    1691--1704

    Renka, R J, 1988, `Multivariate interpolation of large sets of
    scattered data`, ACM Trans. Math. Software (14), 139--148

    Renka, R J, 1988, `Algorithm 661: QSHEP3D: Quadratic Shepard method
    for trivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 151--152

    Shepard, D, 1968, `A two-dimensional interpolation function for
    irregularly spaced data`, Proc. 23rd Nat. Conf. ACM, 517--523,
    Brandon/Systems Press Inc., Princeton
    """

    fun_name = 'naginterfaces.library.interp.dim3_scat_shep'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _liq = _b_utils._EngineIntScalarType(
        2*_m_py+1, {'fun_name': fun_name, 'entity': '_liq'},
    ).to_ctypes()
    _iq = _b_utils._EngineIntArrayType.empty_ndarray(
        (2*_m_py+1,),
    )
    _lrq = _b_utils._EngineIntScalarType(
        10*_m_py+7, {'fun_name': fun_name, 'entity': '_lrq'},
    ).to_ctypes()
    _rq = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (10*_m_py+7,),
    )
    _b_interp.dim3_scat_shep(
        _m,
        _x,
        _y,
        _z,
        _f,
        nw,
        nq,
        _iq,
        _liq,
        _rq,
        _lrq,
    )
    _iq_py = _iq
    _rq_py = _rq
    _return_data = Dim3ScatShepReturnData(
        _iq_py,
        _rq_py,
    )
    return _return_data

Dim3ScatShepEvalReturnData = _collections.namedtuple(
    'Dim3ScatShepEvalReturnData',
    ['q', 'qx', 'qy', 'qz'],
)

def dim3_scat_shep_eval(x, y, z, f, iq, rq, u, v, w):
    u"""
    Interpolated values, evaluate interpolant computed by
    ``dim3_scat_shep``, function and first derivatives, three variables.

    ``dim3_scat_shep_eval`` evaluates the three-dimensional
    interpolating function generated by ``dim3_scat_shep`` and its first
    partial derivatives.

    For full information please refer to the NAG Library document for
    e01th

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01thf.html

    Parameters
    ----------
    x : float, array-like, shape (m)
        m, `x`, `y`, `z` and `f` must be the same values as were
        supplied in the preceding call to ``dim3_scat_shep``.

    y : float, array-like, shape (m)
        m, `x`, `y`, `z` and `f` must be the same values as were
        supplied in the preceding call to ``dim3_scat_shep``.

    z : float, array-like, shape (m)
        m, `x`, `y`, `z` and `f` must be the same values as were
        supplied in the preceding call to ``dim3_scat_shep``.

    f : float, array-like, shape (m)
        m, `x`, `y`, `z` and `f` must be the same values as were
        supplied in the preceding call to ``dim3_scat_shep``.

    iq : int, array-like, shape (2*m+1)
        Must be unchanged from the value returned from a previous call
        to ``dim3_scat_shep``.

    rq : float, array-like, shape (10*m+7)
        Must be unchanged from the value returned from a previous call
        to ``dim3_scat_shep``.

    u : float, array-like, shape (n)
        `u`[i-1], `v`[i-1], `w`[i-1] must be set to the evaluation point
        (u_i,v_i,w_i), for i = 1,2,...,n.

    v : float, array-like, shape (n)
        `u`[i-1], `v`[i-1], `w`[i-1] must be set to the evaluation point
        (u_i,v_i,w_i), for i = 1,2,...,n.

    w : float, array-like, shape (n)
        `u`[i-1], `v`[i-1], `w`[i-1] must be set to the evaluation point
        (u_i,v_i,w_i), for i = 1,2,...,n.

    Returns
    -------
    q : float, ndarray, shape (n)
        `q`[i-1] contains the value of the interpolant, at
        (u_i,v_i,w_i), for i = 1,2,...,n. If any of these evaluation
        points lie outside the region of definition of the interpolant
        the corresponding entries in `q` are set to an extrapolated
        approximation, and ``dim3_scat_shep_eval`` returns with `errno`
        = 3.

    qx : float, ndarray, shape (n)
        `qx`[i-1], `qy`[i-1], `qz`[i-1] contains the value of the
        partial derivatives of the interpolant Q(x,y,z) at
        (u_i,v_i,w_i), for i = 1,2,...,n. If any of these evaluation
        points lie outside the region of definition of the interpolant,
        the corresponding entries in `qx`, `qy` and `qz` are set to
        extrapolated approximations to the partial derivatives, and
        ``dim3_scat_shep_eval`` returns with `errno` = 3.

    qy : float, ndarray, shape (n)
        `qx`[i-1], `qy`[i-1], `qz`[i-1] contains the value of the
        partial derivatives of the interpolant Q(x,y,z) at
        (u_i,v_i,w_i), for i = 1,2,...,n. If any of these evaluation
        points lie outside the region of definition of the interpolant,
        the corresponding entries in `qx`, `qy` and `qz` are set to
        extrapolated approximations to the partial derivatives, and
        ``dim3_scat_shep_eval`` returns with `errno` = 3.

    qz : float, ndarray, shape (n)
        `qx`[i-1], `qy`[i-1], `qz`[i-1] contains the value of the
        partial derivatives of the interpolant Q(x,y,z) at
        (u_i,v_i,w_i), for i = 1,2,...,n. If any of these evaluation
        points lie outside the region of definition of the interpolant,
        the corresponding entries in `qx`, `qy` and `qz` are set to
        extrapolated approximations to the partial derivatives, and
        ``dim3_scat_shep_eval`` returns with `errno` = 3.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, lrq is too small: lrq = *<value>*.

        (`errno` 1)
            On entry, liq is too small: liq = *<value>*.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 10.

        (`errno` 2)
            On entry, values in `rq` appear to be invalid. Check that
            `rq` has not been corrupted between calls to
            ``dim3_scat_shep`` and ``dim3_scat_shep_eval``.

        (`errno` 2)
            On entry, values in `iq` appear to be invalid. Check that
            `iq` has not been corrupted between calls to
            ``dim3_scat_shep`` and ``dim3_scat_shep_eval``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, at least one evaluation point lies outside the
            region of definition of the interpolant. At such points the
            corresponding values in `q` and `qx` contain extrapolated
            approximations. Points should be evaluated one by one to
            identify extrapolated values.

    Notes
    -----
    ``dim3_scat_shep_eval`` takes as input the interpolant Q(x,y,z) of a
    set of scattered data points (x_r,y_r,z_r,f_r), for r = 1,2,...,m,
    as computed by ``dim3_scat_shep``, and evaluates the interpolant and
    its first partial derivatives at the set of points (u_i,v_i,w_i),
    for i = 1,2,...,n.

    ``dim3_scat_shep_eval`` must only be called after a call to
    ``dim3_scat_shep``.

    This function is derived from the function QS3GRD described by Renka
    (1988).

    References
    ----------
    Renka, R J, 1988, `Algorithm 661: QSHEP3D: Quadratic Shepard method
    for trivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 151--152
    """

    fun_name = 'naginterfaces.library.interp.dim3_scat_shep_eval'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _iq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iq'},
        iq,
        copy=False,
    )
    _liq_py = _b_utils._EngineIntArrayType.get_shape(_iq, {'fun_name': fun_name, 'entity': 'iq'}, exp_rank=1)[0]
    _liq = _b_utils._EngineIntScalarType(
        _liq_py, {'fun_name': fun_name, 'entity': '_liq'},
    ).to_ctypes()
    _rq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rq'},
        rq,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lrq_py = _b_utils._EngineFloat64ArrayType.get_shape(_rq, {'fun_name': fun_name, 'entity': 'rq'}, exp_rank=1)[0]
    _lrq = _b_utils._EngineIntScalarType(
        _lrq_py, {'fun_name': fun_name, 'entity': '_lrq'},
    ).to_ctypes()
    _u = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'u'},
        u,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _v = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'v'},
        v,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _w = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _q = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _qx = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _qy = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _qz = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_interp.dim3_scat_shep_eval(
            _m,
            _x,
            _y,
            _z,
            _f,
            _iq,
            _liq,
            _rq,
            _lrq,
            _n,
            _u,
            _v,
            _w,
            _q,
            _qx,
            _qy,
            _qz,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _q_py = _q
    _qx_py = _qx
    _qy_py = _qy
    _qz_py = _qz
    _return_data = Dim3ScatShepEvalReturnData(
        _q_py,
        _qx_py,
        _qy_py,
        _qz_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

Dim4ScatShepReturnData = _collections.namedtuple(
    'Dim4ScatShepReturnData',
    ['iq', 'rq'],
)

def dim4_scat_shep(x, f, nw, nq):
    u"""
    Interpolating functions, modified Shepard's method, four variables.

    ``dim4_scat_shep`` generates a four-dimensional interpolant to a set
    of scattered data points, using a modified Shepard method.

    For full information please refer to the NAG Library document for
    e01tk

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01tkf.html

    Parameters
    ----------
    x : float, array-like, shape (4, m)
        `x`[0,r-1],...,`x`[3,r-1] must be set to the Cartesian
        coordinates of the data point x_r, for r = 1,2,...,m.

    f : float, array-like, shape (m)
        `f`[r-1] must be set to the data value f_r, for r = 1,2,...,m.

    nw : int
        The number N_w of data points that determines each radius of
        influence R_w, appearing in the definition of each of the
        weights w_r, for r = 1,2,...,m (see [Notes]). Note that R_w is
        different for each weight. If `nw` <= 0 the default value `nw` =
        min(32, m-1) is used instead.

    nq : int
        The number N_q of data points to be used in the least squares
        fit for coefficients defining the quadratic functions q_r(x)
        (see [Notes]). If `nq` <= 0 the default value `nq` = min(38,
        m-1) is used instead.

    Returns
    -------
    iq : int, ndarray, shape (2*m+1)
        Integer data defining the interpolant Q(x).

    rq : float, ndarray, shape (15*m+9)
        Real data defining the interpolant Q(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nw` = *<value>* and m = *<value>*.

            Constraint: `nw` <= min(50, m-1).

        (`errno` 1)
            On entry, `nq` = *<value>* and m = *<value>*.

            Constraint: `nq` <= min(50, m-1).

        (`errno` 1)
            On entry, `nq` = *<value>*.

            Constraint: `nq` <= 0 or `nq` >= 14.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 16.

        (`errno` 2)
            There are duplicate nodes in the dataset. `x`[i-1,k-1] =
            `x`[j-1,k-1], for i = *<value>*, j = *<value>* and k =
            1,2,...,4. The interpolant cannot be derived.

        (`errno` 3)
            On entry, all the data points lie on the same
            three-dimensional hypersurface. No unique solution exists.

    Notes
    -----
    ``dim4_scat_shep`` constructs a smooth function Q(x), x in R^4 which
    interpolates a set of m scattered data points (x_r,f_r), for r =
    1,2,...,m, using a modification of Shepard's method.
    The surface is continuous and has continuous first partial
    derivatives.

    The basic Shepard method, which is a generalization of the
    two-dimensional method described in Shepard (1968), interpolates the
    input data with the weighted mean

        Q(x) = sum_r = 1^mw_r(x)q_r/(sum_r = 1^mw_r(x)),

    where q_r = f_r, w_r(x) = 1/d_r^2 and d_r^2 = ||x-x_r||_2^2.

    The basic method is global in that the interpolated value at any
    point depends on all the data, but ``dim4_scat_shep`` uses a
    modification (see Franke and Nielson (1980) and Renka (1988a)),
    whereby the method becomes local by adjusting each w_r(x) to be zero
    outside a hypersphere with centre x_r and some radius R_w.
    Also, to improve the performance of the basic method, each q_r above
    is replaced by a function q_r(x), which is a quadratic fitted by
    weighted least squares to data local to x_r and forced to
    interpolate (x_r,f_r).
    In this context, a point x is defined to be local to another point
    if it lies within some distance R_q of it.

    The efficiency of ``dim4_scat_shep`` is enhanced by using a cell
    method for nearest neighbour searching due to Bentley and Friedman
    (1979) with a cell density of 3.

    The radii R_w and R_q are chosen to be just large enough to include
    N_w and N_q data points, respectively, for user-supplied constants
    N_w and N_q.
    Default values of these arguments are provided by the function, and
    advice on alternatives is given in [Further Comments].

    ``dim4_scat_shep`` is derived from the new implementation of QSHEP3
    described by Renka (1988b).
    It uses the modification for high-dimensional interpolation
    described by Berry and Minser (1999).

    Values of the interpolant Q(x) generated by ``dim4_scat_shep``, and
    its first partial derivatives, can subsequently be evaluated for
    points in the domain of the data by a call to
    ``dim4_scat_shep_eval``.

    References
    ----------
    Bentley, J L and Friedman, J H, 1979, `Data structures for range
    searching`, ACM Comput. Surv. (11), 397--409

    Berry, M W, Minser, K S, 1999, `Algorithm 798: high-dimensional
    interpolation using the modified Shepard method`, ACM Trans. Math.
    Software (25), 353--366

    Franke, R and Nielson, G, 1980, `Smooth interpolation of large sets
    of scattered data`, Internat. J. Num. Methods Engrg. (15),
    1691--1704

    Renka, R J, 1988, `Multivariate interpolation of large sets of
    scattered data`, ACM Trans. Math. Software (14), 139--148

    Renka, R J, 1988, `Algorithm 661: QSHEP3D: Quadratic Shepard method
    for trivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 151--152

    Shepard, D, 1968, `A two-dimensional interpolation function for
    irregularly spaced data`, Proc. 23rd Nat. Conf. ACM, 517--523,
    Brandon/Systems Press Inc., Princeton
    """

    fun_name = 'naginterfaces.library.interp.dim4_scat_shep'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _iq = _b_utils._EngineIntArrayType.empty_ndarray(
        (2*_m_py+1,),
    )
    _rq = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (15*_m_py+9,),
    )
    _b_interp.dim4_scat_shep(
        _m,
        _x,
        _f,
        nw,
        nq,
        _iq,
        _rq,
    )
    _iq_py = _iq
    _rq_py = _rq
    _return_data = Dim4ScatShepReturnData(
        _iq_py,
        _rq_py,
    )
    return _return_data

Dim4ScatShepEvalReturnData = _collections.namedtuple(
    'Dim4ScatShepEvalReturnData',
    ['q', 'qx'],
)

def dim4_scat_shep_eval(x, f, iq, rq, xe):
    u"""
    Interpolated values, evaluate interpolant computed by
    ``dim4_scat_shep``, function and first derivatives, four variables.

    ``dim4_scat_shep_eval`` evaluates the four-dimensional interpolating
    function generated by ``dim4_scat_shep`` and its first partial
    derivatives.

    For full information please refer to the NAG Library document for
    e01tl

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01tlf.html

    Parameters
    ----------
    x : float, array-like, shape (4, m)
        Note: the coordinates of x_r are stored in
        `x`[0,r-1]...`x`[3,r-1].

        **Must** be the same array supplied as argument x in the
        preceding call to ``dim4_scat_shep``. It **must** remain
        unchanged between calls.

    f : float, array-like, shape (m)
        **Must** be the same array supplied as argument f in the
        preceding call to ``dim4_scat_shep``. It **must** remain
        unchanged between calls.

    iq : int, array-like, shape (2*m+1)
        **Must** be the same array returned as argument iq in the
        preceding call to ``dim4_scat_shep``. It **must** remain
        unchanged between calls.

    rq : float, array-like, shape (15*m+9)
        **Must** be the same array returned as argument rq in the
        preceding call to ``dim4_scat_shep``. It **must** remain
        unchanged between calls.

    xe : float, array-like, shape (4, n)
        `xe`[0,r-1],...,`xe`[3,r-1] must be set to the evaluation point
        x_i, for i = 1,2,...,n.

    Returns
    -------
    q : float, ndarray, shape (n)
        `q`[i-1] contains the value of the interpolant, at x_i, for i =
        1,2,...,n. If any of these evaluation points lie outside the
        region of definition of the interpolant the corresponding
        entries in `q` are set to an extrapolated approximation, and
        ``dim4_scat_shep_eval`` returns with `errno` = 3.

    qx : float, ndarray, shape (4, n)
        `qx`[j-1,i-1] contains the value of the partial derivatives with
        respect to x_j of the interpolant Q(x) at x_i, for i =
        1,2,...,n, and for each of the four partial derivatives j =
        1,2,3,4. If any of these evaluation points lie outside the
        region of definition of the interpolant, the corresponding
        entries in `qx` are set to extrapolated approximations to the
        partial derivatives, and ``dim4_scat_shep_eval`` returns with
        `errno` = 3.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 16.

        (`errno` 2)
            On entry, values in `rq` appear to be invalid. Check that
            `rq` has not been corrupted between calls to
            ``dim4_scat_shep`` and ``dim4_scat_shep_eval``.

        (`errno` 2)
            On entry, values in `iq` appear to be invalid. Check that
            `iq` has not been corrupted between calls to
            ``dim4_scat_shep`` and ``dim4_scat_shep_eval``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, at least one evaluation point lies outside the
            region of definition of the interpolant. At such points the
            corresponding values in `q` and `qx` contain extrapolated
            approximations. Points should be evaluated one by one to
            identify extrapolated values.

    Notes
    -----
    ``dim4_scat_shep_eval`` takes as input the interpolant Q(x), x in
    R^4 of a set of scattered data points (x_r,f_r), for r = 1,2,...,m,
    as computed by ``dim4_scat_shep``, and evaluates the interpolant and
    its first partial derivatives at the set of points x_i, for i =
    1,2,...,n.

    ``dim4_scat_shep_eval`` must only be called after a call to
    ``dim4_scat_shep``.

    ``dim4_scat_shep_eval`` is derived from the new implementation of
    QS3GRD described by Renka (1988).
    It uses the modification for high-dimensional interpolation
    described by Berry and Minser (1999).

    References
    ----------
    Berry, M W, Minser, K S, 1999, `Algorithm 798: high-dimensional
    interpolation using the modified Shepard method`, ACM Trans. Math.
    Software (25), 353--366

    Renka, R J, 1988, `Algorithm 661: QSHEP3D: Quadratic Shepard method
    for trivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 151--152
    """

    fun_name = 'naginterfaces.library.interp.dim4_scat_shep_eval'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xe = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xe'},
        xe,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'), (_xe, 'xe')),
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _iq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iq'},
        iq,
        copy=False,
    )
    _rq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rq'},
        rq,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_xe, {'fun_name': fun_name, 'entity': 'xe'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _q = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _qx = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4, _n_py),
        sorder=_m_sorder,
    )
    try:
        _b_interp.dim4_scat_shep_eval(
            _m,
            _x,
            _f,
            _iq,
            _rq,
            _n,
            _xe,
            _q,
            _qx,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _q_py = _q
    _qx_py = _qx
    _return_data = Dim4ScatShepEvalReturnData(
        _q_py,
        _qx_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

Dim5ScatShepReturnData = _collections.namedtuple(
    'Dim5ScatShepReturnData',
    ['iq', 'rq'],
)

def dim5_scat_shep(x, f, nw, nq):
    u"""
    Interpolating functions, modified Shepard's method, five variables.

    ``dim5_scat_shep`` generates a five-dimensional interpolant to a set
    of scattered data points, using a modified Shepard method.

    For full information please refer to the NAG Library document for
    e01tm

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01tmf.html

    Parameters
    ----------
    x : float, array-like, shape (5, m)
        `x`[0,r-1],...,`x`[4,r-1] must be set to the Cartesian
        coordinates of the data point x_r, for r = 1,2,...,m.

    f : float, array-like, shape (m)
        `f`[r-1] must be set to the data value f_r, for r = 1,2,...,m.

    nw : int
        The number N_w of data points that determines each radius of
        influence R_w, appearing in the definition of each of the
        weights w_r, for r = 1,2,...,m (see [Notes]). Note that R_w is
        different for each weight. If `nw` <= 0 the default value `nw` =
        min(32, m-1) is used instead.

    nq : int
        The number N_q of data points to be used in the least squares
        fit for coefficients defining the quadratic functions q_r(x)
        (see [Notes]). If `nq` <= 0 the default value `nq` = min(50,
        m-1) is used instead.

    Returns
    -------
    iq : int, ndarray, shape (2*m+1)
        Integer data defining the interpolant Q(x).

    rq : float, ndarray, shape (21*m+11)
        Real data defining the interpolant Q(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nw` = *<value>* and m = *<value>*.

            Constraint: `nw` <= min(50, m-1).

        (`errno` 1)
            On entry, `nq` = *<value>* and m = *<value>*.

            Constraint: `nq` <= min(70, m-1).

        (`errno` 1)
            On entry, `nq` = *<value>*.

            Constraint: `nq` <= 0 or `nq` >= 20.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 23.

        (`errno` 2)
            There are duplicate nodes in the dataset. `x`[i-1,k-1] =
            `x`[j-1,k-1], for i = *<value>*, j = *<value>* and k =
            1,2,...,5. The interpolant cannot be derived.

        (`errno` 3)
            On entry, all the data points lie on the same
            four-dimensional hypersurface. No unique solution exists.

    Notes
    -----
    ``dim5_scat_shep`` constructs a smooth function Q(x), x in R^5 which
    interpolates a set of m scattered data points (x_r,f_r), for r =
    1,2,...,m, using a modification of Shepard's method.
    The surface is continuous and has continuous first partial
    derivatives.

    The basic Shepard method, which is a generalization of the
    two-dimensional method described in Shepard (1968), interpolates the
    input data with the weighted mean

        Q(x) = sum_r = 1^mw_r(x)q_r/(sum_r = 1^mw_r(x)),

    where q_r = f_r, w_r(x) = 1/d_r^2 and d_r^2 = ||x-x_r||_2^2.

    The basic method is global in that the interpolated value at any
    point depends on all the data, but ``dim5_scat_shep`` uses a
    modification (see Franke and Nielson (1980) and Renka (1988a)),
    whereby the method becomes local by adjusting each w_r(x) to be zero
    outside a hypersphere with centre x_r and some radius R_w.
    Also, to improve the performance of the basic method, each q_r above
    is replaced by a function q_r(x), which is a quadratic fitted by
    weighted least squares to data local to x_r and forced to
    interpolate (x_r,f_r).
    In this context, a point x is defined to be local to another point
    if it lies within some distance R_q of it.

    The efficiency of ``dim5_scat_shep`` is enhanced by using a cell
    method for nearest neighbour searching due to Bentley and Friedman
    (1979) with a cell density of 3.

    The radii R_w and R_q are chosen to be just large enough to include
    N_w and N_q data points, respectively, for user-supplied constants
    N_w and N_q.
    Default values of these arguments are provided, and advice on
    alternatives is given in [Further Comments].

    ``dim5_scat_shep`` is derived from the new implementation of QSHEP3
    described by Renka (1988b).
    It uses the modification for five-dimensional interpolation
    described by Berry and Minser (1999).

    Values of the interpolant Q(x) generated by ``dim5_scat_shep``, and
    its first partial derivatives, can subsequently be evaluated for
    points in the domain of the data by a call to
    ``dim5_scat_shep_eval``.

    References
    ----------
    Bentley, J L and Friedman, J H, 1979, `Data structures for range
    searching`, ACM Comput. Surv. (11), 397--409

    Berry, M W, Minser, K S, 1999, `Algorithm 798: high-dimensional
    interpolation using the modified Shepard method`, ACM Trans. Math.
    Software (25), 353--366

    Franke, R and Nielson, G, 1980, `Smooth interpolation of large sets
    of scattered data`, Internat. J. Num. Methods Engrg. (15),
    1691--1704

    Renka, R J, 1988, `Multivariate interpolation of large sets of
    scattered data`, ACM Trans. Math. Software (14), 139--148

    Renka, R J, 1988, `Algorithm 661: QSHEP3D: Quadratic Shepard method
    for trivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 151--152

    Shepard, D, 1968, `A two-dimensional interpolation function for
    irregularly spaced data`, Proc. 23rd Nat. Conf. ACM, 517--523,
    Brandon/Systems Press Inc., Princeton
    """

    fun_name = 'naginterfaces.library.interp.dim5_scat_shep'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _iq = _b_utils._EngineIntArrayType.empty_ndarray(
        (2*_m_py+1,),
    )
    _rq = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (21*_m_py+11,),
    )
    _b_interp.dim5_scat_shep(
        _m,
        _x,
        _f,
        nw,
        nq,
        _iq,
        _rq,
    )
    _iq_py = _iq
    _rq_py = _rq
    _return_data = Dim5ScatShepReturnData(
        _iq_py,
        _rq_py,
    )
    return _return_data

Dim5ScatShepEvalReturnData = _collections.namedtuple(
    'Dim5ScatShepEvalReturnData',
    ['q', 'qx'],
)

def dim5_scat_shep_eval(x, f, iq, rq, xe):
    u"""
    Interpolated values, evaluate interpolant computed by
    ``dim5_scat_shep``, function and first derivatives, five variables.

    ``dim5_scat_shep_eval`` evaluates the five-dimensional interpolating
    function generated by ``dim5_scat_shep`` and its first partial
    derivatives.

    For full information please refer to the NAG Library document for
    e01tn

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01tnf.html

    Parameters
    ----------
    x : float, array-like, shape (5, m)
        **Must** be the same array supplied as argument x in the
        preceding call to ``dim5_scat_shep``. It **must** remain
        unchanged between calls.

    f : float, array-like, shape (m)
        **Must** be the same array supplied as argument f in the
        preceding call to ``dim5_scat_shep``. It **must** remain
        unchanged between calls.

    iq : int, array-like, shape (2*m+1)
        **Must** be the same array returned as argument iq in the
        preceding call to ``dim5_scat_shep``. It **must** remain
        unchanged between calls.

    rq : float, array-like, shape (21*m+11)
        **Must** be the same array returned as argument rq in the
        preceding call to ``dim5_scat_shep``. It **must** remain
        unchanged between calls.

    xe : float, array-like, shape (5, n)
        `xe`[0,i-1],...,`xe`[4,i-1] must be set to the evaluation point
        x_i, for i = 1,2,...,n.

    Returns
    -------
    q : float, ndarray, shape (n)
        `q`[i-1] contains the value of the interpolant, at x_i, for i =
        1,2,...,n. If any of these evaluation points lie outside the
        region of definition of the interpolant the corresponding
        entries in `q` are set to an extrapolated approximation, and
        ``dim5_scat_shep_eval`` returns with `errno` = 3.

    qx : float, ndarray, shape (5, n)
        `qx`[j-1,i-1] contains the value of the partial derivatives with
        respect to x_j of the interpolant Q(x) at x_i, for i =
        1,2,...,n, and for each of the five partial derivatives j =
        1,2,3,4,5. If any of these evaluation points lie outside the
        region of definition of the interpolant, the corresponding
        entries in `qx` are set to extrapolated approximations to the
        partial derivatives, and ``dim5_scat_shep_eval`` returns with
        `errno` = 3.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 23.

        (`errno` 2)
            On entry, values in `rq` appear to be invalid. Check that
            `rq` has not been corrupted between calls to
            ``dim5_scat_shep`` and ``dim5_scat_shep_eval``.

        (`errno` 2)
            On entry, values in `iq` appear to be invalid. Check that
            `iq` has not been corrupted between calls to
            ``dim5_scat_shep`` and ``dim5_scat_shep_eval``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, at least one evaluation point lies outside the
            region of definition of the interpolant. At such points the
            corresponding values in `q` and `qx` contain extrapolated
            approximations. Points should be evaluated one by one to
            identify extrapolated values.

    Notes
    -----
    ``dim5_scat_shep_eval`` takes as input the interpolant Q(x), x in
    R^5 of a set of scattered data points (x_r,f_r), for r = 1,2,...,m,
    as computed by ``dim5_scat_shep``, and evaluates the interpolant and
    its first partial derivatives at the set of points x_i, for i =
    1,2,...,n.

    ``dim5_scat_shep_eval`` must only be called after a call to
    ``dim5_scat_shep``.

    ``dim5_scat_shep_eval`` is derived from the new implementation of
    QS3GRD described by Renka (1988).
    It uses the modification for five-dimensional interpolation
    described by Berry and Minser (1999).

    References
    ----------
    Berry, M W, Minser, K S, 1999, `Algorithm 798: high-dimensional
    interpolation using the modified Shepard method`, ACM Trans. Math.
    Software (25), 353--366

    Renka, R J, 1988, `Algorithm 661: QSHEP3D: Quadratic Shepard method
    for trivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 151--152
    """

    fun_name = 'naginterfaces.library.interp.dim5_scat_shep_eval'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xe = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xe'},
        xe,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'), (_xe, 'xe')),
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _iq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iq'},
        iq,
        copy=False,
    )
    _rq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rq'},
        rq,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_xe, {'fun_name': fun_name, 'entity': 'xe'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _q = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _qx = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (5, _n_py),
        sorder=_m_sorder,
    )
    try:
        _b_interp.dim5_scat_shep_eval(
            _m,
            _x,
            _f,
            _iq,
            _rq,
            _n,
            _xe,
            _q,
            _qx,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _q_py = _q
    _qx_py = _qx
    _return_data = Dim5ScatShepEvalReturnData(
        _q_py,
        _qx_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

DimnGridReturnData = _collections.namedtuple(
    'DimnGridReturnData',
    ['ans'],
)

def dimn_grid(narr, uniform, axis, v, point, method, k=1, wf=0.0):
    u"""
    Interpolates an n-dimensional point on grid data using either
    linear, cubic or modified Shepard's method.

    ``dimn_grid`` interpolates data at a point in n-dimensional space,
    that is defined by a set of gridded data points.
    It offers three methods to interpolate the data: Linear
    Interpolation, Cubic Interpolation and Weighted Average.

    For full information please refer to the NAG Library document for
    e01za

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01zaf.html

    Parameters
    ----------
    narr : int, array-like, shape (d)
        The number of data ordinates in each dimension, with `narr`[i-1]
        = n_i, for i = 1,2,...,d.

    uniform : bool
        States whether the data points are uniformly spaced.

        `uniform` = True

            The data points are uniformly spaced.

        `uniform` = False

            The data points are not uniformly spaced.

    axis : float, array-like, shape (lx)
        Defines the axis. If the data points are uniformly spaced (see
        argument `uniform`) `axis` should contain the start and end of
        each dimension (x_{1 1},x_{1 n_1},...,x_{d 1},x_{d n_d}).

        If the data points are not uniformly spaced, `axis` should
        contain all the data ordinates for each dimension
        (x_{1 1},x_{1 2},...,x_{1 n_1},...,x_{d 1},x_{d 2},...,x_{d n_d}).

    v : float, array-like, shape (sum(`narr`))
        Holds the values of the data points in such an order that the
        index of a data value with coordinates (z_1,z_2,...,z_d) is

            sum(z_iprod_{n in S_i}n),

        where S_i = {`narr`[l-1]:l = 1,...,i-1} e.g.,
        ((x_{11},x_{21},...,x_{d1}),(x_{12},x_{21},...,x_{d1}),...,(x_{1n_d},x_{21},...,x_{d1}),(x_{11},x_{22},...,x_{d1}),(x_{12},x_{22},...,x_{d1}),...,(x_{1n_d},x_{2n_d},...,x_{dn_d})).

    point : float, array-like, shape (d)
        x, the point at which the data value is to be interpolated.

    method : int
        The method to be used.

        `method` = 1

            Weighted Average.

        `method` = 2

            Linear Interpolation.

        `method` = 3

            Cubic Interpolation.

    k : int, optional
        If `method` = 1, `k` controls the number of data points used in
        the Weighted Average method, with `k` points used in each
        dimension, either side of the interpolation point. The total
        number of data points used for the interpolation will,
        therefore, be (2*`k`)^d.

        If `method` != 1, then `k` is not referenced and need not be
        set.

    wf : float, optional
        The power used for the weighted average such that a high power
        will cause closer points to be more heavily weighted.

    Returns
    -------
    ans : float
        Holds the result of the interpolation.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, d = *<value>*.

            Constraint: d >= 2.

        (`errno` 2)
            On entry, `narr`[*<value>*] = *<value>*.

            Constraint: `narr`[i-1] >= 2.

        (`errno` 4)
            On entry, `axis` decreases in dimension *<value>*.

            Constraint: `axis` definition must be strictly increasing.

        (`errno` 5)
            On entry, lx = *<value>*, sum of `narr`= *<value>*.

            Constraint: if `uniform` = False, lx = sum of `narr`.

        (`errno` 5)
            On entry, lx = *<value>*, d = *<value>*.

            Constraint: if `uniform` = True, lx = 2*d.

        (`errno` 7)
            On entry, `k` = *<value>*.

            Constraint: if `method` = 1, `k` >= 1.

        (`errno` 8)
            On entry, `point`[*<value>*] = *<value>* and data range =
            [*<value>*,*<value>*].

            Constraint: `point` must be within the data range.

        (`errno` 9)
            On entry, `method` = 3 and `uniform` = False.

            Constraint: if `method` = 3, `uniform` must be True.

        (`errno` 9)
            On entry, `method` = *<value>*.

            Constraint: `method` = 1, 2 or 3.

        (`errno` 10)
            On entry, `wf` = *<value>*.

            Constraint: if `method` = 1, 1.0 <= `wf` <= 15.0.

        (`errno` 101)
            Cubic Interpolation method does not have enough data
            surrounding `point`; interpolation not possible.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 201)
            Warning: the size of `k` has been reduced, due to too few
            data points around `point`.

    Notes
    -----
    ``dimn_grid`` interpolates an n-dimensional point within a set of
    gridded data points, Z = {z_{1j_1},z_{2j_2},...,z_{dj_d}}, with
    corresponding data values F = {f_{1j_1},f_{2j_2},...,f_{dj_d}}
    where, for the ith dimension, j_i = 1,...,n_i and n_i is the number
    of ordinates in the ith dimension.

    A hypercube of (2k)^d data points [h_1,h_2,...,h_(2k)^d]⊂Z, where
    h_i = (h_{i1},h_{i2},...,h_{id}) and the corresponding data values
    are f(h_i) in F, around the given point, x = (x_1,x_2,...,x_d), is
    found and then used to interpolate using one of the following three
    methods.

    (i) Weighted Average, that is a modification of Shepard's method
        (Shepard (1968)) as used for scattered data in
        ``dimn_scat_shep``. This method interpolates the data with the
        weighted mean

            Q(x) = sum_r = 1^(2k)^dw_r(x)f_r/(sum_r = 1^(2k)^dw_r(x)),

        where f_r = f(h_r), w_r(x) = 1/D(|x-h_r|) and D(y) =
        y_1^rho+y_2^rho+...+y_d^rho, for a given value of rho.

    (#) Linear Interpolation, which takes 2^d surrounding data points (k
        = 1) and performs two one-dimensional linear interpolations in
        each dimension on data points h_a and h_b, reducing the
        dimension every iteration until it has reached an answer. The
        formula for linear interpolation in dimension i is simply

            f = f_a+(x_i-h_{ai})(f_b-f_a)/(h_{bi}-h_{ai}),

        where f_r = f(h_r) and h_{ai} < x_i < h_{bi}.

    (#) Cubic Interpolation, based on cubic convolution (Keys (1981)).
        In a similar way to the Linear Interpolation method, it performs
        the interpolations in one dimension reducing it each time,
        however it requires four surrounding data points in each
        dimension (k = 2), two in each direction (h_-1,h_0,h_1,h_2). The
        following is used to calculate the one-dimensional interpolant
        in dimension i

            f = 1/2(1 t t^2 t^3)[table omitted](f_-1;f_0;f_1;f_2)

        where t = x_i-h_{0i} and f_r = f(h_r).

    References
    ----------
    Keys, R, 1981, `Cubic Convolution Interpolation for Digital Image
    Processing`, IEEE Transactions on Acoutstics, Speech, and Signal
    Processing (Vol ASSP-29 No. 6), 1153--1160,
    http://hmi.stanford.edu/doc/Tech_Notes/HMI-TN-2004-004-Interpolation/Keys_cubic_interp.pdf

    Shepard, D, 1968, `A two-dimensional interpolation function for
    irregularly spaced data`, Proc. 23rd Nat. Conf. ACM, 517--523,
    Brandon/Systems Press Inc., Princeton
    """

    fun_name = 'naginterfaces.library.interp.dimn_grid'

    _narr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'narr'},
        narr,
        copy=False,
    )
    _d_py = _b_utils._EngineIntArrayType.get_shape(_narr, {'fun_name': fun_name, 'entity': 'narr'}, exp_rank=1)[0]
    _d = _b_utils._EngineIntScalarType(
        _d_py, {'fun_name': fun_name, 'entity': '_d'},
    ).to_ctypes()
    _axis = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'axis'},
        axis,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lx_py = _b_utils._EngineFloat64ArrayType.get_shape(_axis, {'fun_name': fun_name, 'entity': 'axis'}, exp_rank=1)[0]
    _lx = _b_utils._EngineIntScalarType(
        _lx_py, {'fun_name': fun_name, 'entity': '_lx'},
    ).to_ctypes()
    _v = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'v'},
        v,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _point = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'point'},
        point,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalar = _b_interp.dimn_grid(
            _d,
            _narr,
            uniform,
            _axis,
            _lx,
            _v,
            _point,
            method,
            k,
            wf,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _ans_py = _return_scalar
    _return_data = DimnGridReturnData(
        _ans_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ans_py

DimnScatShepReturnData = _collections.namedtuple(
    'DimnScatShepReturnData',
    ['iq', 'rq'],
)

def dimn_scat_shep(x, f, nw=-1, nq=-1):
    u"""
    Interpolating function, modified Shepard's method, d dimensions.

    ``dimn_scat_shep`` generates a multidimensional interpolant to a set
    of scattered data points, using a modified Shepard method.
    When the number of dimensions is no more than five, there are
    corresponding functions in submodule ``interp`` which are specific
    to the given dimensionality. ``dim2_scat_shep`` generates the
    two-dimensional interpolant, while ``dim3_scat_shep``,
    ``dim4_scat_shep`` and ``dim5_scat_shep`` generate the three-, four
    - and five-dimensional interpolants respectively.

    For full information please refer to the NAG Library document for
    e01zm

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01zmf.html

    Parameters
    ----------
    x : float, array-like, shape (d, m)
        The d components of the first data point must be stored in
        elements 0,1,...,d-1 of `x`. The second data point must be
        stored in elements d,d+1,...,2*d-1 of `x`, and so on. In
        general, the m data points must be stored in `x`[j,i], for j =
        0,1,...,d-1, for i = 0,1,...,m-1.

    f : float, array-like, shape (m)
        `f`[r-1] must be set to the data value f_r, for r = 1,2,...,m.

    nw : int, optional
        The number N_w of data points that determines each radius of
        influence R_w, appearing in the definition of each of the
        weights w_r, for r = 1,2,...,m (see [Notes]). Note that R_w is
        different for each weight. If `nw` <= 0 the default value `nw` =
        min(2*(d+1)*(d+2), m-1) is used instead.

    nq : int, optional
        The number N_q of data points to be used in the least squares
        fit for coefficients defining the quadratic functions q_r(x)
        (see [Notes]). If `nq` <= 0 the default value `nq` =
        min((d+1)*(d+2)*6/5, m-1) is used instead.

    Returns
    -------
    iq : int, ndarray, shape (2*m+1)
        Integer data defining the interpolant Q(x).

    rq : float, ndarray, shape (lrq)
        Real data defining the interpolant Q(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nw` = *<value>* and m = *<value>*.

            Constraint: `nw` <= m-1.

        (`errno` 1)
            On entry, `nq` = *<value>* and m = *<value>*.

            Constraint: `nq` <= m-1.

        (`errno` 1)
            On entry, `nq` = *<value>* and d = *<value>*.

            Constraint: `nq` <= 0 or `nq` >= (d+1)*(d+2)/2-1.

        (`errno` 1)
            On entry, m = *<value>* and d = *<value>*.

            Constraint: m >= (d+1)*(d+2)/2+2.

        (`errno` 1)
            On entry, ((d+1)*(d+2)/2)*m+2*d+1 exceeds the largest
            machine integer.

            d = *<value>* and m = *<value>*.

        (`errno` 1)
            On entry, d = *<value>*.

            Constraint: d >= 2.

        (`errno` 2)
            There are duplicate nodes in the dataset. `x`[k-1,i-1] =
            `x`[k-1,j-1], for i = *<value>*, j = *<value>* and k =
            1,2,...,d. The interpolant cannot be derived.

        (`errno` 3)
            On entry, all the data points lie on the same hypersurface.
            No unique solution exists.

    Notes
    -----
    ``dimn_scat_shep`` constructs a smooth function Q(x), x in R^d which
    interpolates a set of m scattered data points (x_r,f_r), for r =
    1,2,...,m, using a modification of Shepard's method.
    The surface is continuous and has continuous first partial
    derivatives.

    The basic Shepard method, which is a generalization of the
    two-dimensional method described in Shepard (1968), interpolates the
    input data with the weighted mean

        Q(x) = sum_r = 1^mw_r(x)q_r/(sum_r = 1^mw_r(x)),

    where q_r = f_r, w_r(x) = 1/(||x-x_r||_2^2).

    The basic method is global in that the interpolated value at any
    point depends on all the data, but ``dimn_scat_shep`` uses a
    modification (see Franke and Nielson (1980) and Renka (1988a)),
    whereby the method becomes local by adjusting each w_r(x) to be zero
    outside a hypersphere with centre x_r and some radius R_w.
    Also, to improve the performance of the basic method, each q_r above
    is replaced by a function q_r(x), which is a quadratic fitted by
    weighted least squares to data local to x_r and forced to
    interpolate (x_r,f_r).
    In this context, a point x is defined to be local to another point
    if it lies within some distance R_q of it.

    The efficiency of ``dimn_scat_shep`` is enhanced by using a cell
    method for nearest neighbour searching due to Bentley and Friedman
    (1979) with a cell density of 3.

    The radii R_w and R_q are chosen to be just large enough to include
    N_w and N_q data points, respectively, for user-supplied constants
    N_w and N_q.
    Default values of these parameters are provided, and advice on
    alternatives is given in [Further Comments].

    ``dimn_scat_shep`` is derived from the new implementation of QSHEP3
    described by Renka (1988b).
    It uses the modification for high-dimensional interpolation
    described by Berry and Minser (1999).

    Values of the interpolant Q(x) generated by ``dimn_scat_shep``, and
    its first partial derivatives, can subsequently be evaluated for
    points in the domain of the data by a call to
    ``dimn_scat_shep_eval``.

    References
    ----------
    Bentley, J L and Friedman, J H, 1979, `Data structures for range
    searching`, ACM Comput. Surv. (11), 397--409

    Berry, M W, Minser, K S, 1999, `Algorithm 798: high-dimensional
    interpolation using the modified Shepard method`, ACM Trans. Math.
    Software (25), 353--366

    Franke, R and Nielson, G, 1980, `Smooth interpolation of large sets
    of scattered data`, Internat. J. Num. Methods Engrg. (15),
    1691--1704

    Renka, R J, 1988, `Multivariate interpolation of large sets of
    scattered data`, ACM Trans. Math. Software (14), 139--148

    Renka, R J, 1988, `Algorithm 661: QSHEP3D: Quadratic Shepard method
    for trivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 151--152

    Shepard, D, 1968, `A two-dimensional interpolation function for
    irregularly spaced data`, Proc. 23rd Nat. Conf. ACM, 517--523,
    Brandon/Systems Press Inc., Princeton
    """

    fun_name = 'naginterfaces.library.interp.dimn_scat_shep'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _d = _b_utils._EngineIntScalarType(
        _d_py, {'fun_name': fun_name, 'entity': '_d'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _iq = _b_utils._EngineIntArrayType.empty_ndarray(
        (2*_m_py+1,),
    )
    if _m_py == 0:
        _vl_lrq_py = 0
    elif _d_py <= int(float((sqrt(float(_m_py)/float(2))*sqrt(float(_b_machine.integer_max())/float(2))-1))/float(float(_m_py)/float(2))):
        _vl_lrq_py = ((_d_py+1)*(_d_py+2)//2)*_m_py+2*_d_py+1
    else:
        _vl_lrq_py = 0
    _rq = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_vl_lrq_py,),
    )
    _b_interp.dimn_scat_shep(
        _d,
        _m,
        _x,
        _f,
        nw,
        nq,
        _iq,
        _rq,
    )
    _iq_py = _iq
    _rq_py = _rq
    _return_data = DimnScatShepReturnData(
        _iq_py,
        _rq_py,
    )
    return _return_data

DimnScatShepEvalReturnData = _collections.namedtuple(
    'DimnScatShepEvalReturnData',
    ['q', 'qx'],
)

def dimn_scat_shep_eval(x, f, iq, rq, xe):
    u"""
    Interpolated values, evaluate interpolant computed by
    ``dimn_scat_shep``, function and first derivatives, d dimensions.

    ``dimn_scat_shep_eval`` evaluates the multidimensional interpolating
    function generated by ``dimn_scat_shep`` and its first partial
    derivatives.

    For full information please refer to the NAG Library document for
    e01zn

    https://www.nag.com/numeric/nl/nagdoc_28.7/flhtml/e01/e01znf.html

    Parameters
    ----------
    x : float, array-like, shape (d, m)
        Note: the ith ordinate of the point x_j is stored in
        `x`[i-1,j-1].

        **Must** be the same array supplied as argument x in the
        preceding call to ``dimn_scat_shep``. It **must** remain
        unchanged between calls.

    f : float, array-like, shape (m)
        **Must** be the same array supplied as argument f in the
        preceding call to ``dimn_scat_shep``. It **must** remain
        unchanged between calls.

    iq : int, array-like, shape (2*m+1)
        **Must** be the same array returned as argument iq in the
        preceding call to ``dimn_scat_shep``. It **must** remain
        unchanged between calls.

    rq : float, array-like, shape (lrq)
        **Must** be the same array returned as argument rq in the
        preceding call to ``dimn_scat_shep``. It **must** remain
        unchanged between calls.

    xe : float, array-like, shape (d, n)
        Note: the ith ordinate of the point x_j is stored in
        `xe`[i-1,j-1].

        `xe`[0,j-1],...,`xe`[d-1,j-1] must be set to the evaluation
        point x_j, for j = 1,2,...,n.

    Returns
    -------
    q : float, ndarray, shape (n)
        `q`[i-1] contains the value of the interpolant, at x_i, for i =
        1,2,...,n. If any of these evaluation points lie outside the
        region of definition of the interpolant the corresponding
        entries in `q` are set to an extrapolated approximation, and
        ``dimn_scat_shep_eval`` returns with `errno` = 3.

    qx : float, ndarray, shape (d, n)
        `qx`[i-1,j-1] contains the value of the partial derivatives with
        respect to the ith independent variable (dimension) of the
        interpolant Q(x) at x_j, for j = 1,2,...,n, and for each of the
        partial derivatives i = 1,2,...,d. If any of these evaluation
        points lie outside the region of definition of the interpolant,
        the corresponding entries in `qx` are set to extrapolated
        approximations to the partial derivatives, and
        ``dimn_scat_shep_eval`` returns with `errno` = 3.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, m = *<value>* and d = *<value>*.

            Constraint: m >= (d+1)*(d+2)/2+2.

        (`errno` 1)
            On entry, ((d+1)*(d+2)/2)*m+2*d+1 exceeds the largest
            machine integer.

            d = *<value>* and m = *<value>*.

        (`errno` 1)
            On entry, d = *<value>*.

            Constraint: d >= 2.

        (`errno` 2)
            On entry, values in `rq` appear to be invalid. Check that
            `rq` has not been corrupted between calls to
            ``dimn_scat_shep`` and ``dimn_scat_shep_eval``.

        (`errno` 2)
            On entry, values in `iq` appear to be invalid. Check that
            `iq` has not been corrupted between calls to
            ``dimn_scat_shep`` and ``dimn_scat_shep_eval``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, at least one evaluation point lies outside the
            region of definition of the interpolant. At such points the
            corresponding values in `q` and `qx` contain extrapolated
            approximations. Points should be evaluated one by one to
            identify extrapolated values.

    Notes
    -----
    ``dimn_scat_shep_eval`` takes as input the interpolant Q(x), x in
    R^d of a set of scattered data points (x_r,f_r), for r = 1,2,...,m,
    as computed by ``dimn_scat_shep``, and evaluates the interpolant and
    its first partial derivatives at the set of points x_i, for i =
    1,2,...,n.

    ``dimn_scat_shep_eval`` must only be called after a call to
    ``dimn_scat_shep``.

    ``dimn_scat_shep_eval`` is derived from the new implementation of
    QS3GRD described by Renka (1988).
    It uses the modification for high-dimensional interpolation
    described by Berry and Minser (1999).

    References
    ----------
    Berry, M W, Minser, K S, 1999, `Algorithm 798: high-dimensional
    interpolation using the modified Shepard method`, ACM Trans. Math.
    Software (25), 353--366

    Renka, R J, 1988, `Algorithm 661: QSHEP3D: Quadratic Shepard method
    for trivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 151--152
    """

    fun_name = 'naginterfaces.library.interp.dimn_scat_shep_eval'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xe = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xe'},
        xe,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'), (_xe, 'xe')),
    )
    _d_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _d = _b_utils._EngineIntScalarType(
        _d_py, {'fun_name': fun_name, 'entity': '_d'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _iq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iq'},
        iq,
        copy=False,
    )
    if _m_py == 0:
        _vl_lrq_py = 0
    elif _d_py <= int(float((sqrt(float(_m_py)/float(2))*sqrt(float(_b_machine.integer_max())/float(2))-1))/float(float(_m_py)/float(2))):
        _vl_lrq_py = ((_d_py+1)*(_d_py+2)//2)*_m_py+2*_d_py+1
    else:
        _vl_lrq_py = 0
    _rq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rq'},
        rq,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_xe, {'fun_name': fun_name, 'entity': 'xe'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _q = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _qx = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_d_py, _n_py),
        sorder=_m_sorder,
    )
    try:
        _b_interp.dimn_scat_shep_eval(
            _d,
            _m,
            _x,
            _f,
            _iq,
            _rq,
            _n,
            _xe,
            _q,
            _qx,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _q_py = _q
    _qx_py = _qx
    _return_data = DimnScatShepEvalReturnData(
        _q_py,
        _qx_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data
